/*
This program is the main file of the Magnetic Disk simulator. All parameters need to be set here.
The design of the magnetic disk simulator is as follows:
UserBits -> Encoder -> Channel -> LPF -> Downsampler -> Equalizer -> Viterbi(SOVA) -> Decoder
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <time.h>
#include "MagneticDisk.h"

#define PI 3.1415926

#define LARGE_SPOT 1
#define MEDIUM_SPOT 0

struct Mag_Param{
	double sigma_t;		/* Sigma of T-Profile in nm */
	double T_Peak;		/* Peak Tempearture in centigrade */
	double Orig_T_Peak;	/* Same as T_Peak. Used when temperature varying effects are included */
	double c;			/* Position of peak temperature along downtrack w.r.t gap centreline in nm */
	double d;			/* Position of peak temperature along crosstrack w.r.t track centre in nm*/
	double Orig_d;		/* Same as d. Used when write head cross track movement effects are included */
	double z0;			/* Cross track location. T_calc = T_peak * exp(-(z0-d)^2/(2*sigma_t^2)) */
	double Hc[2];		/* Dependence of coercivity on Temperature in A/m. Hc = aT + b => {a,b} */
	double Mr[2];		/* Dependence of remnant magnetization on Temperature in A/m. Mr = aT + b => {a,b} */
	double S[2];		/* Dependence of coercive squareness on Temperature. S = aT + b => {a,b} */
	double Hg;			/* Deep gap field in A/m*/
};

struct Physical_Param{
	float g;			/* Write/Head gap in nm*/
	float d;			/* Medium spacing(in nm) or the space between the bottom of pole and the surface of the medium */
	float t;			/* Medium thickness in nm */
	float y;			/* Distance from bottom of pole to the centre of the medium in nm*/
	float wt;			/* Width of the track */
};

struct Reader_Param{
	double C;			/* Constant which describes the physical attributes of GMR reader */
	double gr;			/* Spacing between the sheild and the GMR element (nm)*/
	double tr;			/* Width of the GMR element (nm)*/
	double wr;			/* Width of the reader (in Z direction, i.e. crosstrack direction) */
	double sigma_r;		/* Sigma of Reader sensitivity function, which is Gaussian (nm)*/
};

/*------------------------- ERROR-FUNCTION RELATED FUNCTIONS ------------------------*/

/* Adapted from "Numerical Recipes in C" */
#define ITMAX 100
#define EPS1 3.0e-7
#define FPMIN 1.0e-30

float gammln(float xx)
{
// Returns the value ln[(xx)] for xx > 0.
// Internal arithmetic will be done in double precision, a nicety that you can omit if five-figure
// accuracy is good enough.
	double x,y,tmp,ser;
	static double cof[6]={76.18009172947146,-86.50532032941677,
	24.01409824083091,-1.231739572450155,
	0.1208650973866179e-2,-0.5395239384953e-5};
	int j;
	y=x=xx;
	tmp=x+5.5;
	tmp -= (x+0.5)*log(tmp);
	ser=1.000000000190015;
	for (j=0;j<=5;j++)
		ser += cof[j]/++y;
	return -tmp+log(2.5066282746310005*ser/x);
}

void gser(float *gamser, float a, float x, float *gln)
{
	// Returns the incomplete gamma function P(a, x) evaluated by its series representation as gamser.
	// Also returns ln (a) as gln.
	float gammln(float xx);
	//void nrerror(char error_text[]);
	int n;
	float sum,del,ap;
	*gln=gammln(a);
	if (x <= 0.0) {
		if (x < 0.0) printf("%s\n","x less than 0 in routine gser");
		*gamser=0.0;
		return;
	} else {
		ap=a;
		del=sum=1.0/a;
		for (n=1;n<=ITMAX;n++) {
			++ap;
			del *= x/ap;
			sum += del;
			if (fabs(del) < fabs(sum)*EPS1) {
				*gamser=sum*exp(-x+a*log(x)-(*gln));
				return;
			}
		}
		printf("%s\n","a too large, ITMAX too small in routine gser");
		return;
	}
}


void gcf(float *gammcf, float a, float x, float *gln)
{
// Returns the incomplete gamma function Q(a, x) evaluated by its continued fraction representation
// as gammcf. Also returns ln(a) as gln.
	float gammln(float xx);
	//void nrerror(char error_text[]);
	int i;
	float an,b,c,d,del,h;
	*gln=gammln(a);
	b=x+1.0-a; 			// Set up for evaluating continued fraction
						// by modified Lentzs method (5.2)
						// with b0 = 0.
	c=1.0/FPMIN;
	d=1.0/b;
	h=d;
	for (i=1;i<=ITMAX;i++) 	//Iterate to convergence.
	{
		an = -i*(i-a);
		b += 2.0;
		d=an*d+b;
		if (fabs(d) < FPMIN) d=FPMIN;
		c=b+an/c;
		if (fabs(c) < FPMIN) c=FPMIN;
		d=1.0/d;
		del=d*c;
		h *= del;
		if (fabs(del-1.0) < EPS1) break;
	}
	if (i > ITMAX) printf("%s\n","a too large, ITMAX too small in gcf");
	*gammcf=exp(-x+a*log(x)-(*gln))*h; 			// Put factors in front.
	return;
}

float gammp(float a, float x)
//Returns the incomplete gamma function P(a, x).
{
	void gcf(float *gammcf, float a, float x, float *gln);
	void gser(float *gamser, float a, float x, float *gln);
	//void nrerror(char error_text[]);
	float gamser,gammcf,gln;
	if (x < 0.0 || a <= 0.0)
		printf("%s\n","Invalid arguments in routine gammp");
	if (x < (a+1.0)) 	// Use the series representation.
	{
		gser(&gamser,a,x,&gln);
		return gamser;
	} else { 			// Use the continued fraction representation
		gcf(&gammcf,a,x,&gln);
		return 1.0-gammcf;	// and take its complement.
	}
}

double erf(double x)
{
// Returns the error function erf(x).
	float gammp(float a, float x);
	return x < 0.0 ? -gammp(0.5,x*x) : gammp(0.5,x*x);
}

/*---------------------- UNIFORM RANDOM NUMBER GENERATOR ------------------------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGenerator(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}


/*---------------------- UNIFORM RANDOM NUMBER GENERATOR FOR PULSE ------------------------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGeneratorPulse(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}


/*---------------------- UNIFORM RANDOM NUMBER GENERATOR FOR JITTER ------------------------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGeneratorJitter(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}


/*---------------------- UNIFORM RANDOM NUMBER GENERATOR FOR USER BITS ------------------------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGeneratorUserBits(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}



/*----------- UNIFORM RANDOM NUMBER GENERATOR FOR WRITE HEAD CROSS TRACK MOVEMENT ---------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGeneratorWrHdCrTrMov(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}


/*------------------------- UNIFORM RANDOM NUMBER GENERATOR FOR EQUALIZER ------------------------*/

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-07
#define RNMX (1.0-EPS)

float UniformRandomNumberGeneratorEqualizer(long *idum)
{
 int j=0;
 long k=0;
 static long iy=0;
 static long iv[NTAB];
 float temp=0;
 	if(*idum<=0 || !iy)			/* Initialize*/
		{
 		 if (-(*idum)<1)
			*idum=1;
		 else
			*idum=-(*idum);     /* Be sure to prevent idum=0*/
 		 for(j=NTAB+7;j>=0;j--)
			{
			 k=(*idum)/IQ;
			 *idum=IA*(*idum-k*IQ)-IR*k;
			 if(*idum<0)
				*idum+=IM;
			 if(j<NTAB)
				iv[j]=*idum;
			}
		 iy=iv[0];
		}
 	k=(*idum)/IQ;				/* Start here when not initializing */
	*idum=IA*(*idum-k*IQ)-IR*k; /* Compute idum=(IA*idum)%IM without overflows*/
 	if(*idum<0)
		*idum+=IM;

	j=iy/NDIV;
	iy=iv[j]; 					/* Will be in the range 0..NTAB-1 */
	iv[j]=*idum;
 	if((temp=AM*iy) >RNMX)
		return RNMX;  			/* Because users dont expect endpoint values */
	else
		return temp;
}


/*--------------------------- GAUSSIAN RANDOM NUMBER GENERATOR -------------------------*/

 /* The function returns a normally distributed deviate with zero mean and unit variance, using
   uniform_rand (idum)as the source of uniform deviates.
    Reference: Numerical Recipes in C
	      Chapter 7, Random Numbers (pg. no. 289)
    typed by : Sundararajan Sankaranarayanan
   date     : 2nd January, 2001 */
   float GaussianRandomNumberGenerator(long *idum)
 {
  static int iset=0;
  static float gset;
  float fac=0;
  float rsq=0;
  float v1,v2=0;
 	if(iset==0)
		{
		 do
			{ /* pick tow uniform numbers in the square extending from -1 to +1 in
			     each direction, see if they are in the unit circle, and if they
			     are not, try again. */
 			 v1=2.0*UniformRandomNumberGenerator(idum)-1.0;
			 v2=2.0*UniformRandomNumberGenerator(idum)-1.0;
			 rsq=v1*v1+v2*v2;
			} while (rsq>=1.0 || rsq == 0.0);
		 fac=sqrt(-2.0*log(rsq)/rsq);
		 gset=v1*fac;
		 iset=1;  /* Set flag */
		 return v2*fac;
		}
	else
		{ /* We have an extra deviate handy. so unset the flag and return it */
		 iset=0;
		 return gset;
		}
}

/*------------- GAUSSIAN RANDOM NUMBER GENERATOR FOR TEMPERATURE VARIATION --------------*/

 /* The function returns a normally distributed deviate with zero mean and unit variance, using
   uniform_rand (idum)as the source of uniform deviates.
    Reference: Numerical Recipes in C
	      Chapter 7, Random Numbers (pg. no. 289)
    typed by : Sundararajan Sankaranarayanan
   date     : 2nd January, 2001 */
  float GaussianRandomNumberGeneratorTemp(long *idum)
 {
  static int iset=0;
  static float gset;
  float fac=0;
  float rsq=0;
  float v1,v2=0;
 	if(iset==0)
		{
		 do
			{ /* pick tow uniform numbers in the square extending from -1 to +1 in
			     each direction, see if they are in the unit circle, and if they
			     are not, try again. */
 			 v1=2.0*UniformRandomNumberGenerator(idum)-1.0;
			 v2=2.0*UniformRandomNumberGenerator(idum)-1.0;
			 rsq=v1*v1+v2*v2;
			} while (rsq>=1.0 || rsq == 0.0);
		 fac=sqrt(-2.0*log(rsq)/rsq);
		 gset=v1*fac;
		 iset=1;  /* Set flag */
		 return v2*fac;
		}
	else
		{ /* We have an extra deviate handy. so unset the flag and return it */
		 iset=0;
		 return gset;
		}
}

/*------------------- GAUSSIAN RANDOM NUMBER GENERATOR FOR HMD -------------------------*/

 /* The function returns a normally distributed deviate with zero mean and unit variance, using
   uniform_rand (idum)as the source of uniform deviates.
    Reference: Numerical Recipes in C
	      Chapter 7, Random Numbers (pg. no. 289)
    typed by : Sundararajan Sankaranarayanan
   date     : 2nd January, 2001 */
float GaussianRandomNumberGeneratorHMD(long *idum)
 {
  static int iset=0;
  static float gset;
  float fac=0;
  float rsq=0;
  float v1,v2=0;
 	if(iset==0)
		{
		 do
			{ /* pick tow uniform numbers in the square extending from -1 to +1 in
			     each direction, see if they are in the unit circle, and if they
			     are not, try again. */
 			 v1=2.0*UniformRandomNumberGeneratorJitter(idum)-1.0;
			 v2=2.0*UniformRandomNumberGeneratorJitter(idum)-1.0;
			 rsq=v1*v1+v2*v2;
			} while (rsq>=1.0 || rsq == 0.0);
		 fac=sqrt(-2.0*log(rsq)/rsq);
		 gset=v1*fac;
		 iset=1;  /* Set flag */
		 return v2*fac;
		}
	else
		{ /* We have an extra deviate handy. so unset the flag and return it */
		 iset=0;
		 return gset;
		}
}

/*------------------- GAUSSIAN RANDOM NUMBER GENERATOR FOR JITTER NOISE -------------------------*/

 /* The function returns a normally distributed deviate with zero mean and unit variance, using
   uniform_rand (idum)as the source of uniform deviates.
    Reference: Numerical Recipes in C
	      Chapter 7, Random Numbers (pg. no. 289)
    typed by : Sundararajan Sankaranarayanan
   date     : 2nd January, 2001 */
float GaussianRandomNumberGeneratorJitter(long *idum)
 {
  static int iset=0;
  static float gset;
  float fac=0;
  float rsq=0;
  float v1,v2=0;
 	if(iset==0)
		{
		 do
			{ /* pick tow uniform numbers in the square extending from -1 to +1 in
			     each direction, see if they are in the unit circle, and if they
			     are not, try again. */
 			 v1=2.0*UniformRandomNumberGeneratorJitter(idum)-1.0;
			 v2=2.0*UniformRandomNumberGeneratorJitter(idum)-1.0;
			 rsq=v1*v1+v2*v2;
			} while (rsq>=1.0 || rsq == 0.0);
		 fac=sqrt(-2.0*log(rsq)/rsq);
		 gset=v1*fac;
		 iset=1;  /* Set flag */
		 return v2*fac;
		}
	else
		{ /* We have an extra deviate handy. so unset the flag and return it */
		 iset=0;
		 return gset;
		}
}


/*------------- GAUSSIAN RANDOM NUMBER GENERATOR FOR PULSE BROADENING EFFECT ---------------*/

 /* The function returns a normally distributed deviate with zero mean and unit variance, using
   uniform_rand (idum)as the source of uniform deviates.
    Reference: Numerical Recipes in C
	      Chapter 7, Random Numbers (pg. no. 289)
    typed by : Sundararajan Sankaranarayanan
   date     : 2nd January, 2001 */
float GaussianRandomNumberGeneratorPulse(long *idum)
 {
  static int iset=0;
  static float gset;
  float fac=0;
  float rsq=0;
  float v1,v2=0;
 	if(iset==0)
		{
		 do
			{ /* pick tow uniform numbers in the square extending from -1 to +1 in
			     each direction, see if they are in the unit circle, and if they
			     are not, try again. */
 			 v1=2.0*UniformRandomNumberGeneratorPulse(idum)-1.0;
			 v2=2.0*UniformRandomNumberGeneratorPulse(idum)-1.0;
			 rsq=v1*v1+v2*v2;
			} while (rsq>=1.0 || rsq == 0.0);
		 fac=sqrt(-2.0*log(rsq)/rsq);
		 gset=v1*fac;
		 iset=1;  /* Set flag */
		 return v2*fac;
		}
	else
		{ /* We have an extra deviate handy. so unset the flag and return it */
		 iset=0;
		 return gset;
		}
}


/*-------------------------------------- MATRIX MULTIPLICATION ---------------------------------------*/

/*
Matrix Multiplication: C = A*B
A is of size [m * n]
B is of size [p * q]
C is the output matrix of size [m * q]
*/

void MatrixMult(double **A, int m, int n,double **B, int p, int q, double **C)
{
	int i,j,k;

	/* If the dimensions do not match, terminate the program */
	if(n!=p)
	{
		printf("\nERROR: Matrix A and B cannot be multiplied. Program terminating\n");
		exit(-1);
	}

	/* Memory for matrix C should have already been allocated and intialized to 0 */
	for(i=0;i<m;i++)
	{
		for(j=0;j<q;j++)
		{
			for(k=0;k<n;k++)
			{
				C[i][j] += A[i][k]*B[k][j];
			}
		}
	}
}


/*-------------------------------------- MATRIX INVERSE ---------------------------------------*/

/*
Calculates inverse of matrix A and stores in matrix B. Assumes that Matrix A is FULL RANK.
Inverse (i.e. B) is calculated by row-reducing matrix A to an identity matrix and simultaneously
applying the same row operations to B, which initially is an identity matrix.
*/

void MatrixInv(double **TempA, double **B, int m){
	int r,c,i,j;
	double temp;
	double **A;

	/* Copying the input matrix TempA into matrix A */
	A = (double**)malloc(m*sizeof(double));
	for(i=0;i<m;i++)
	{
		A[i] = (double*)malloc(m*sizeof(double));
	}

	for(i=0;i<m;i++)
	{
		for(j=0;j<m;j++)
		{
			A[i][j] = TempA[i][j];
		}
	}

	/* Creating the identity matrix B */
	/* Memory for Matrix B should already be allocated and initialized to 0 */
	for(i=0;i<m;i++)
	{
			B[i][i] = 1;
	} 

	/*** Performing Gaussian elimination on A ***/

	for(c=0;c<m;c++)
	{/* loop over all columns of A */

		/*
		If diagnol element of column c is 0, then interchange row c with any other row
		so that diagnol element of column c is non-zero.
		*/
		if(A[c][c] == 0)
		{
			for(i=0;i<m;i++)
			{
				if(A[i][c] != 0)
				{
					/** Interchange row i and row c **/
					for(j=0;j<m;j++)
					{
						temp = A[i][j];
						A[i][j] = A[c][j];
						A[c][j] = temp;
						/* Do the same operation for B */
						temp = B[i][j];
						B[i][j] = B[c][j];
						B[c][j] = temp;
					}
					break;
				}
			}
		}

		/*
		Making the diagnol element of column c, A[c][c] 1 by dividing
		all the elements of row c, i.e. A[c][0 1 ... m-1] by A[c][c].
		*/
		if(A[c][c] != 1)
		{
			temp = A[c][c];
			for(j=0;j<m;j++)
			{
				A[c][j] = A[c][j]/temp;
				/* Do the same operation on the matrix B */
				B[c][j] = B[c][j]/temp;
			}
		}

		for(i=0;i<m;i++)
		{/* loop over all rows of A */

			/*
			Making all elements of column c as zero, except for the diagnol element,
			which is 1 at this point.
			*/
			if(i!=c)
			{
				temp = A[i][c];
				for(j=0;j<m;j++)
				{
					A[i][j] = A[i][j] - temp*A[c][j];
					B[i][j] = B[i][j] - temp*B[c][j];
				}
			}
		}/* loop over all rows of A */

	}/* loop over all columns of A */
	
	for(i=0;i<m;i++)
	{
		free(A[i]);
	}
	free(A);
}

/*-------------------------------------- CORRELATION ---------------------------------------*/

/*
Calculates the correlation of A and B for a fixed delay.
r_ab{delta} = E{A_k * B_(k-delta)}
*/

void Corr(double *A, int LengthA, double *B, int LengthB, int shift, double *output)
{
	int k,start,finish;

	/** Determinimg the index for which array A and shifted version of array B will overlap **/
	/* B is shifted to the right w.r.t A */
	if(shift >= 0)
	{
		start = shift;
	}else
	/* B is shifted to the left w.r.t A */
	{
		start = 0;
	}
	if(LengthA - 1 <= shift + LengthB - 1)
	{
		finish = LengthA - 1;
	}else
	{
		finish = shift + LengthB - 1;
	}

	/* Return if there is no overlap between A and shifted version of B */
	if(start > finish)
	{
		output[0] = 0;
		return;
	}

	/* Calculate the correlation */
	output[0] = 0.0;
	for(k=start;k<=finish;k++)
	{
		output[0] += A[k]*B[k-shift];
	}
}


/*-------------------------------------- GPR TARGET ---------------------------------------*/

/*
Calculates the Generalized Partial Response target for a given length.
For the algorithm and variable notations, refer to Jaekyun Moon et. al.
'Equalization for Maximum Likelihood Detectors', IEEE Trans. Mag., Vol 31, No. 2, Mar 1995.
*/
#define PRE 1e-05

void FindGPRTarget(double *s, double *a, double *F, double *G, int SeqLength, int L, int N)
{
	int i,j;
	int K;
	double **A,**R,**T;
	double Lambda;
	double **RInv, **TTranspose, **Temp1,**Temp2;

	/*
	s: channel output sequence
	a: bipolar input to the channel
	F: Equalizer non-causal filter taps (length N)
	G: optimum GPR target	(length L)
	N: filter tap length = 2*K + 1. The taps are {f(-K)....f(0)....f(+K)}.

	A: [L*L] autocorrelation matrix of a
	R: [N*N] autocorrelation matrix of s
	T: [N*L] cross correlation matrix of s and a
	*/

	R = (double**)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		R[i] = (double*)malloc(N*sizeof(double));
	}

	A = (double**)malloc(L*sizeof(double));
	for(i=0;i<L;i++)
	{
		A[i] = (double*)malloc(L*sizeof(double));
	}

	T = (double**)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		T[i] = (double*)malloc(L*sizeof(double));
	}

	K = floor(N/2.0);	/* Will work for both N even and odd */

	/** Calculate R **/
	/*
	Calculating first row of R. Others can be derived from this since autocorrelation
	matrix is symmetric.
	*/
	for(i=0;i<N;i++)
	{
		Corr(s, SeqLength, s, SeqLength, i, &(R[0][i]));
	}

	for(i=1;i<N;i++)
	{
		for(j=0;j<N;j++)
		{
			R[i][j] = R[0][abs(j-i)];
		}
	}

	/** Calculate A **/
	/*
	Calculating first row of A. Others can be derived from this since autocorrelation
	matrix is symmetric.
	*/
	for(i=0;i<L;i++)
	{
		Corr(a, SeqLength, a, SeqLength, i, &(A[0][i]));
	}

	for(i=1;i<L;i++)
	{
		for(j=0;j<L;j++)
		{
			A[i][j] = A[0][abs(j-i)];
		}
	}

	/** Calculate T **/
	for(i=0;i<N;i++)
	{
		for(j=0;j<L;j++)
		{
			Corr(s, SeqLength, a, SeqLength, ((K-i) + j), &(T[i][j]));
		}
	}

	/** Calculate Lambda **/
	/* Lambda = (I'* (1/(A - T'* 1/R * T)) * I) */
	/* Calculate Inverse of R */
	RInv = (double **)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		RInv[i] = (double*)malloc(N*sizeof(double));
	}
	for(i=0;i<N;i++)
	{
		for(j=0;j<N;j++)
		{
			RInv[i][j] = 0.0;
		}
	}
	MatrixInv(R, RInv, N);

	/** Calculate (RInv * T) **/
	/* Temp1 is where the result of (RInv * T) is stored */
	Temp1 = (double **)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		Temp1[i] = (double*)malloc(L*sizeof(double));
	}
	for(i=0;i<N;i++)
	{
		for(j=0;j<L;j++)
		{
			Temp1[i][j] = 0.0;
		}
	}
	MatrixMult(RInv, N, N, T, N, L, Temp1);

	/** Calculate T' **/
	/* TTranspose is where the result of T' is stored */
	TTranspose = (double **)malloc(L*sizeof(double));
	for(i=0;i<L;i++)
	{
		TTranspose[i] = (double*)malloc(N*sizeof(double));
	}
	for(i=0;i<L;i++)
	{
		for(j=0;j<N;j++)
		{
			TTranspose[i][j] = T[j][i];
		}
	}

	/** Calculate T'* Temp1 **/
	/* Temp2 is where the result of T'* Temp1 is stored */
	Temp2 = (double **)malloc(L*sizeof(double));
	for(i=0;i<L;i++)
	{
		Temp2[i] = (double*)malloc(L*sizeof(double));
	}
	for(i=0;i<L;i++)
	{
		for(j=0;j<L;j++)
		{
			Temp2[i][j] = 0.0;
		}
	}
	MatrixMult(TTranspose, L, N, Temp1, N, L, Temp2);

	/** Calculate (A - Temp2) **/
	/* A is overwritten with the result of (A - Temp2) is stored */
	for(i=0;i<L;i++)
	{
		for(j=0;j<L;j++)
		{
			A[i][j] = A[i][j] - Temp2[i][j];
		}
	}

	/** Calculate Inverse of A **/
	/* Note that A was overwritten in the previous step and is not the same as the original A */
	/* Temp2 is where the result of A Inverse is stored */
	/* Memory for Temp2 has already been allocated */
	for(i=0;i<L;i++)
	{
		for(j=0;j<L;j++)
		{
			Temp2[i][j] = 0.0;
		}
	}
	MatrixInv(A, Temp2, L);

	/** Calculate Lambda **/
	/* Lambda is just the inverse of first element of Temp2 */
	Lambda = 1/Temp2[0][0];

	/** Calculate G **/
	/* G is lambda * [1st column of Temp2] */
 	for(i=0;i<L;i++)
	{
		G[i] = Lambda * Temp2[i][0];
	}

	/* Since we have assumed monic constraint, G[0] should be 1 */
	if( fabs(G[0] - 1.0) > PRE )
	{
		printf("\nERROR: GPRTarget() -> Monic constraint (G[0]=1) not satisfied.\n");
		printf("G[0] = %g\n",G[0]);
		exit(-1);
	}

	/** Calculate F **/
	free(Temp1);
	free(Temp2);

	/* Calculate T*G */
	/* Temp2 is where the result of T*G is stored */
	/*
	Eventhough Temp2 is a [N*1] 1D matrix, we allocate a 2D array because the function
	MatrixMult() expects a 2D array. For the same reason we send Temp1 (2D) instead of G.
	*/
	Temp2 = (double **)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		Temp2[i] = (double*)malloc(1*sizeof(double));
		Temp2[i][0]=0.0;
	}
	Temp1 = (double **)malloc(L*sizeof(double));
	for(i=0;i<L;i++)
	{
		Temp1[i] = (double*)malloc(1*sizeof(double));
		Temp1[i][0] = G[i];
	}
	MatrixMult(T, N, L, Temp1, L, 1, Temp2);

	/* Calculate RInv*Temp2 */
	/* Temp1 is where the result of RInv*Temp2 is stored */
	/*
	Eventhough Temp1 is a [N*1] 1D matrix, we allocate a 2D array, because the function
	MatrixMult expects a 2D array (for the same reason we cannot directly use F, which is passed as
	a 1D array to the GPR function).
	*/
	for(i=0;i<L;i++)
	{
		free(Temp1[i]);
	}
	free(Temp1);
	Temp1 = (double **)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		Temp1[i] = (double*)malloc(1*sizeof(double));
		Temp1[i][0] = 0.0;
	}
	MatrixMult(RInv, N, N, Temp2, N, 1, Temp1);

	for(i=0;i<N;i++)
	{
		F[N-1-i] = Temp1[i][0];
	}
	
	/* Freeing	Memory */
	for(i=0;i<N;i++)
	{
		free(R[i]);
		free(T[i]);
		free(RInv[i]);
		free(Temp1[i]);
	}
	free(R);
	free(T);
	free(RInv);
	free(Temp1);
	for(i=0;i<L;i++)
	{
		free(A[i]);
		free(TTranspose[i]);
		free(Temp2[i]);
	}
	free(A);
	free(TTranspose);
	free(Temp2);
}


/*-------------------------------------- NON-CAUSAL FILTER ---------------------------------------*/

/*
Since, channel and GPR is modeled as non-causal filters, it is coded as a separate function and not clubbed
together with causal FIR function. Also note the difference in the way the impulse response vector is specified.

NOTE: The length of output vector is DataLength + floor(NumTaps/2)
*/

void NonCausalFIR(double *Data, int DataLength, double *Output, double *h, int NumTaps){
	int i=0,j=0;
	double *PaddedData;
	int FrontPadLength=0,BackPadLength=0;

	/*
	PaddedData => padded with certain number of zeros in the front and back of Data
	for easy implementation of FIR.
	*/

	/* h(0) is assumed to be at location h(floor(NumTaps/2)) */
	FrontPadLength = floor(NumTaps/2.0);
	BackPadLength = NumTaps-1;
	PaddedData = (double*)malloc((DataLength + FrontPadLength + BackPadLength)*sizeof(double));

	/* PaddedData: < floor(NumTaps/2) zero's >< Data >< (NumTaps-1) zero's > */
	for(i=0;i<(DataLength+FrontPadLength+BackPadLength);i++)
	{
		if(i<=FrontPadLength-1 || i >= FrontPadLength+DataLength)
		{
			PaddedData[i] = 0;
		}else
		{
			PaddedData[i] = Data[i-FrontPadLength];
		}
	}

	/** FIR filtering **/
	/*
	h: ...... h(-2) h(-1) h(0) h(1) h(2)......
	PaddedData: 0000....Data....000000
	First valid output: ...... + h(-2)*0 + h(-1)*0 + h(0)*Data(0) + h(1)*Data(1) +.....
	Last valid output: h(-floor(NumTaps/2))*Data(DataLength-1) + h(-floor(NumTaps/2)+1)*0 + .....
	*/
	for(i=0;i<(DataLength+FrontPadLength);i++)
	{
		Output[i]=0.0;
		for(j=0;j<NumTaps;j++)
		{
			Output[i] += h[j]*PaddedData[i+j];		/*
													Remember that:
													PaddedData[floor(NumTaps/2)]=Data[0]
			                     	   				PaddedData[floor(NumTaps/2)+DataLength-1] = Data[Datalength-1]
			                     	   				*/
		}
	}
	free(PaddedData);
}


/*-------------------------------------- CAUSAL FIR ---------------------------------------*/

/* CAUSAL FIR filter for LPF and Fixed target PR Equalization */
/* NOTE: Unlike in NonCausal filter, output vector here is of length [DataLegnth+NumTaps-1] */
void CausalFIR(double *Data, int DataLength, double *Output, double *h, int NumTaps)
{
	int i=0,j=0;
	int FrontPadLength,BackPadLength;
	double *PaddedData = NULL;
	double ChannelOutput=0.0;

	/** Padding zeros for ease of FIR implementation **/

	/*
	PaddedData => padded with certain number of zeros in the front of Data
	for easy implementation of FIR.
	*/
	FrontPadLength = NumTaps-1;
	BackPadLength = NumTaps-1;
	PaddedData = (double *)malloc((DataLength + FrontPadLength + BackPadLength)*sizeof(double));

	/* PaddedData: < (NumTaps-1) zero's >< Data > */
	for(i=0;i<(DataLength+FrontPadLength+BackPadLength);i++)
	{
		if(i < FrontPadLength || i > (DataLength+FrontPadLength-1) )
		{ 
			PaddedData[i] = 0.0;
		}else
		{
			PaddedData[i] = Data[i-FrontPadLength];
		}
	}

	/** Causal FIR filtering **/
	/*
	h: ...... h(0) h(-1) h(-2) h(-3)...... NOTE the difference from ChannelFilter's impulse response vector def
	PaddedData: 0000....Data
	First valid output: h(0)*Data(0) + h(-1)*0 + .....
	Last valid output: h(0)*0 + h(-1)*0 + ..... + h(-(NumTaps-1))*Data[DataLength-1]
	*/
	for(i=FrontPadLength;i<(DataLength+FrontPadLength+BackPadLength);i++)
	{
		ChannelOutput=0.0;
		for(j=0;j<NumTaps;j++)
		{
			ChannelOutput += h[j]*PaddedData[i-j];	/*
													Remember that:
													PaddedData[NumTaps-1] = Data[0]
													PaddedData[NumTaps+Datalength-2] = Data[Datalength-1]
													*/
		}
		Output[i-FrontPadLength] = ChannelOutput;
	}
	free(PaddedData);
}


/*-------------------------------------- ADAPT EQUALIZER ---------------------------------------*/

#define STEP_LENGTH 0.005
#define MIN_MSE 0.005

int AdaptEqualizer(double *PRImpRes, int PRImpResLength, double *EqCoeff, int NumEqTaps, double *PaddedEncodedBits, double *DSChannelOutput,int SectorLength, int StartFlag)
{
	int i,j;
	double *PRActualOutput, *EqOutputTemp, *PaddedDSChannelOutput;
	long int idum = -300;
	int FrontPadLength,BackPadLength;
	double EqOutput=0.0, MSE = 0.0, EqError= 0.0;
	double LMSError = 0.0;

	if(StartFlag == 1)
	{
		/* Initialize Equalizer coefficients */
		for(i=0;i<NumEqTaps;i++)
		{
			EqCoeff[i] = 0; //UniformRandomNumberGeneratorEqualizer(&idum);
		}
	}

	/** Calculating the actual PR target output **/
	PRActualOutput = (double *)malloc((SectorLength+PRImpResLength-1)*sizeof(double));
	for(i=0;i<SectorLength;i++)
	{
		PRActualOutput[i] = 0.0;
	}
	CausalFIR(PaddedEncodedBits,SectorLength,PRActualOutput,PRImpRes,PRImpResLength);
	
	/** Padding Data for ease of implementation **/
	FrontPadLength = NumEqTaps/2;
	BackPadLength = NumEqTaps-1;

	PaddedDSChannelOutput = (double *)malloc((SectorLength + FrontPadLength + BackPadLength)*sizeof(double));

	/* PaddedData: < (NumEqTaps-1) zero's >< Data > */
	for(i=0;i<(SectorLength+FrontPadLength+BackPadLength);i++)
	{
		if((i<=FrontPadLength-1)||(i>=SectorLength+FrontPadLength))
		{
			PaddedDSChannelOutput[i] = 0;
		}else
		{
			PaddedDSChannelOutput[i] = DSChannelOutput[i-FrontPadLength];
		}

	}

	/** Adapting Equalizer coefficients for the current sector **/
	for(i=0;i<SectorLength;i++)
	{
		EqOutput = 0.0;
		/* Calculating equalizer output for the current symbol */
		for(j=0;j<NumEqTaps;j++)
		{
			EqOutput += EqCoeff[j]*PaddedDSChannelOutput[i+j];	/*
																Remember that:
																PaddedDSChannelOutput[NumEqTaps/2]=DSChannelOutput[0]
																PaddedDSChannelOutput[SectorLength+FrontPadLength-1]=DSChannelOutput[SectorLength-1]
																*/
		}
		/* Adapting equalizer coefficients for the current symbol */
		EqError = EqOutput - PRActualOutput[i];
		LMSError += EqError*EqError;
		for(j=0;j<NumEqTaps;j++)
		{
			EqCoeff[j] -= 2*STEP_LENGTH*EqError*PaddedDSChannelOutput[i+j];
		}
	}

	/** Calculating MSE for the current sector **/
	EqOutputTemp = (double*)malloc((SectorLength+NumEqTaps/2)*sizeof(double));
	for(i=0;i<SectorLength;i++)
	{
		EqOutputTemp[i] = 0.0;
	}

	NonCausalFIR(DSChannelOutput,SectorLength,EqOutputTemp,EqCoeff,NumEqTaps);
	for(i=0;i<SectorLength;i++)
	{
		MSE += pow((PRActualOutput[i] - EqOutputTemp[i]),2);
	}
	printf("MSE = %g\n",MSE);

	/* Stop adapting if MSE is below some acceptable limit */
	if(MSE <= MIN_MSE)
	{
		return(0);
	}
	
	printf("LMS = %g\n",LMSError/SectorLength);
	free(PaddedDSChannelOutput);
	free(PRActualOutput);
	free(EqOutputTemp);

	return(1);
}


/*-------------------------------------- LPF ---------------------------------------*/

/*
Low-Pass filters a given input sequence according to the given filter order and
cut-off. It uses Hamming window.
*/
void LPF(double *ChannelOutput, int ChannelOutputLength, double *LPFChannelOutput, int FilterOrder, double CutOff)
{
	int i;
	double *h,*w;
	double sum=0.0;

	CutOff = CutOff*PI;
	h = (double *)malloc((FilterOrder+1)*sizeof(double));
	w = (double *)malloc((FilterOrder+1)*sizeof(double));

	/** Calculating LPF coefficients **/
	for(i=FilterOrder;i>=0;i--)
	{
		if(i == (double)FilterOrder/2.0)
		{
			h[i] = CutOff/PI;
		}else
		{
			h[i] = sin(CutOff*(i-(double)FilterOrder/2))/(PI*(i-(double)FilterOrder/2));
		}
		/* Multiply with corresponding Hamming window coefficient */
		h[i] = h[i]*(0.54 - 0.46*cos(2*PI*i/(double)FilterOrder));
		sum += h[i];
	}

	/* Normalizing the magnitude at DC to 1 */
	for(i=0;i<FilterOrder+1;i++)
	{
		h[i] = h[i]/sum;
	}

	/** Low Pass Filtering the input data **/
	NonCausalFIR(ChannelOutput, ChannelOutputLength, LPFChannelOutput, h, (FilterOrder+1));
	free(h);
	free(w);
}


/*-------------------------------------- VITERBI DETECTOR ---------------------------------------*/

void Viterbi(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedOutput, double *PRImpRes, int PRImpResLength)
{
	int i,j,k,NumStates;
	double metric0,metric1;
	double sample0,sample1;
	double *PathMetric[2];
	int *Path[2];
	int *VS[2];
	int MinPathMetricIndex;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 (= Order of PR polynomial) */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = pow(2,(PRImpResLength-1));

	Path[0] = (int *)malloc(NumStates*sizeof(int));
	Path[1] = (int *)malloc(NumStates*sizeof(int));
	/* Path[0][i] => valid path of state i in the previous time */
	/* Path[1][i] => valid path of state i in the current time */
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = 0;
		Path[1][i] = 0;
	}

	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}
	
	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

	/**** Main VITERBI loop ****/

	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	bit operations given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			/* All incoming paths for all states are valid */
			{
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/
					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
					PathMetric[1][i] = metric0;
				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
					PathMetric[1][i] = metric1;
				}
			}

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++)
			{
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedOutput[k-(Delay-1)] = (Path[1][MinPathMetricIndex] >> (Delay-1)) & 1;
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			Path[0][i] = Path[1][i];
			Path[1][i] = 0;
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}
		
		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */

	/* Detecting the last 'delay-1' bits */
	for(i=2;i<=Delay;i++)
	{
		DetectedOutput[SectorLength-1-(Delay-i)] = (Path[0][MinPathMetricIndex] >> (Delay-i)) & 1;
	}
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
}


/*----------------------- SOFT-OUTPUT VITERBI ALGORITHM (SOVA) ----------------------*/

void SOVA(int Delay, double *EqualizedChannelOutput, int SectorLength, int *DetectedHardOutput, double *DetectedSoftOutput, double *PRImpRes, int PRImpResLength, double NoiseSigma)
{
	int i,j,k,NumStates;
	double metric0,metric1;
	double sample0,sample1;
	int state0,state1,survivingstate,MinPathMetricIndex;
	double *PathMetric[2];
	int *Path[2];
	int *VS[2];
	double  **ProbWrongDet[2];
	double NoiseVariance = NoiseSigma*NoiseSigma;
	double Delta = 0.0;
	double ProbWrongSurv = 0.0;
	int DifferingBits = 0;
	int VSLengthPrevTime=1,VSLengthCurrentTime=0;

	/* Number of bits in every state = PRImpResLength-1 */
	/* Number of states in trellis = 2^(Order of PR polynomial) */
	NumStates = pow(2,(PRImpResLength-1));

	Path[0] = (int *)malloc(NumStates*sizeof(int));
	Path[1] = (int *)malloc(NumStates*sizeof(int));
	/* Path[0][i] => valid path of state i in the previous time */
	/* Path[1][i] => valid path of state i in the current time */
	/* Note that the numbers 0 and 1 appended to the variables 'metric','sample' and 'state' DONOT indicate current/past time */
	for(i=0;i<NumStates;i++)
	{
		Path[0][i] = 0;
		Path[1][i] = 0;
	}

	PathMetric[0] = (double *)malloc(NumStates * sizeof(double));
	PathMetric[1] = (double *)malloc(NumStates * sizeof(double));
	/* PathMetric[i] => The path metric of the valid output of state i at current time */
	/*
	NOTE that Pathmetric is initialized to 0 in the hope that by the time 'Delay' bits are
	transmitted, all of these are overwritten. If not, then when making a decision on a bit,
	the state for which Pathmetric is still 0 will be chosen, which is wrong. It cannot be
	initialized to a large number to avoid this problem because it is additive.
	*/
	for(i=0;i<NumStates;i++)
	{
		PathMetric[0][i] = 0.0;
		PathMetric[1][i] = 0.0;
	}

	/*
	ProbWrongDet variables are used to store the probabilities that each bit is detected
	wrongly. As before, 0 indicates previous time and 1 indicates current time
	*/
	ProbWrongDet[0] = (double **)malloc(NumStates*sizeof(double));
	ProbWrongDet[1] = (double **)malloc(NumStates*sizeof(double));
	for(i=0;i<NumStates;i++){
		ProbWrongDet[0][i] = (double *)malloc(Delay*sizeof(double));
		ProbWrongDet[1][i] = (double *)malloc(Delay*sizeof(double));
	}
	for(i=0;i<NumStates;i++)
	{
		for(j=0;j<Delay;j++)
		{
			ProbWrongDet[0][i][j] = 0.0;
			ProbWrongDet[1][i][j] = 0.0;
		}
	}

	/* Initially, until time (PRImpResLength-1), all states are not valid */
	/* VS[0] => Valid States in the previous time */
	/* VS[1] => Valid States in the current time */
	VS[0] = (int *)malloc(sizeof(int));
	VS[1] = (int *)malloc(sizeof(int));
	VS[0][0]=0;	/* Assuming that we start with state 0 at time 0 */
	VS[1][0]=0;

	/**** Main SOVA loop ****/
	for(k=0;k<SectorLength;k++)
	{/* Loop over PaddedEqOutput */

		for(i=0;i<NumStates;i++)
		{/* Loop over all states */

			/*** For each state, calculating the path metric of the two incoming paths ***/

			/*
			If i = [i1 i2 ... in] (the array elements are binary) is the current state
			then its corresponding previous states are [0 i1 i2.....in-1] and [1 i1 i2....in-1].
			These are determined by	doing the bit operations as given below.
			*/

			/**
			Calculating the sample value and metric at current time for the transition from
			state [0 i1 i2....in-1]	to state [i1 i2....in].
			Corresponding Previous state is given by: (CurrentState >> 1) => shifting all
			bits to the right by 1 position and appending a ZERO at the left most position.
			**/
			sample0 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample0 += PRImpRes[j]*((i >> j) & 1);
			}
			sample0 += PRImpRes[PRImpResLength-1]*0;	/* Eventhough it is 0, it is explicitly written for clarity */
			sample0 = sample0*2;
			metric0 = PathMetric[0][i >> 1] + pow((EqualizedChannelOutput[k] - sample0),2);

			/**
			Calculating the sample value and metric at current time for the transition from
			state [1 i1 i2....in-1]	to state [i1 i2....in]
			Corresponding Previous state is given by: (CurrentState >> 1) | (1 << (CurrentStateLength-1))
			=> shifting all bits to the right by 1 position and appending a ONE at the
			left most position
			**/
			sample1 = 0.0;
			for(j=0;j<PRImpResLength-1;j++)
			{
				sample1 += PRImpRes[j]*((i >> j) & 1);
			}
			sample1 += PRImpRes[PRImpResLength-1]*1;
			sample1 = sample1*2;
			metric1 = PathMetric[0][(i >> 1)|(1 << (PRImpResLength-2))] + pow((EqualizedChannelOutput[k] - sample1),2);

			state0 = i >> 1;
			state1 = (i >> 1)|(1 << (PRImpResLength-2));

			/* 
			Assuming that we start with state 0, then until time (PRImpResLength-1),
			all states will have less than 2 valid incoming paths. For example, at time 1, the 
			incoming paths for any state is valid only if it comes from state 0, because we 
			assume that we start from state 0 in time 0.
			*/
			if(k < (PRImpResLength-1))
			{
				for(j=0;j<VSLengthPrevTime;j++)
				{
					if((i >> 1)==VS[0][j])
					{	
						/* This is the only valid incoming path */
						Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
						PathMetric[1][i] = metric0;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
					if(((i >> 1) | (1 << (PRImpResLength-2))) == VS[0][j])
					{
						/* This is the only valid incoming path */
						Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
						PathMetric[1][i] = metric1;
						VSLengthCurrentTime++;
						VS[1] = (int*)realloc(VS[1],(VSLengthCurrentTime)*sizeof(int));
						/* 
						This state now becomes valid, i.e. any path emerging out of
						this state is valid in the next time
						*/
						VS[1][VSLengthCurrentTime-1] = i;
						break;	/* At most only one path is valid */
					}
				}
			}
			else
			{ /* Loop: All states and all incoming paths are valid */
				if(metric0 <= metric1)
				{
					/*
					New path is from the state [0 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [0 i1...in-1] in the previous time appended with bit 'in'
					*/

					/*
					(Path[0][i >> 1] << 1) => shifting all bits of path of the state [0 i1...in-1]
					to the left by 1 position.
					(i & 1) => The last bit of state i.
					(Path[0][i >> 1] << 1) | (i & 1) => appending last bit of i to the left-shifted path.
					*/
					Path[1][i] = (Path[0][i >> 1] << 1) | (i & 1);
					PathMetric[1][i] = metric0;
					survivingstate = state0;

				}else{
					/*
					New path is from the state [1 i1 i2...in-1], i.e current valid path for state i is
					the valid path of state [1 i1...in-1] in the previous time appended with bit 'in'
					*/
					/* Bit operations are done as before */
					Path[1][i] = (Path[0][(i >> 1)|(1 << (PRImpResLength-2))] << 1) | (i & 1);
					PathMetric[1][i] = metric1;
					survivingstate = state1;
				}

				/** Update Probability of wrong detection of bits **/
				Delta = fabs(metric0 - metric1)/(2*NoiseVariance);
				ProbWrongSurv = 1/(1+exp(Delta));

				/* Calculate the bits where the two incoming paths to state i differ, using XOR */
				DifferingBits = Path[0][state0] ^ Path[1][state1];

				/* Obviously, they differ at the current time */
				ProbWrongDet[1][i][Delay-1] = ProbWrongSurv;

				/* For all bits, calculate Probablity of wrong detection */
				/*
				Note that we do not consider the bit in the path 'Delay' time period away (i.e 0th element in array),
				because it has already been detected in the previous iteration and will be overwritten in this iteration.
				*/
				for(j=0;j<Delay-1;j++)
				{
					/* Bits between the two paths differ */
					if(((DifferingBits >> j) & 1) == 1)
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j]*(1-ProbWrongSurv) + (1-ProbWrongDet[0][survivingstate][Delay-1-j])*ProbWrongSurv;
					}
					/* Bits between the two paths dont differ */
					else
					{
						ProbWrongDet[1][i][Delay-2-j] = ProbWrongDet[0][survivingstate][Delay-1-j];
					}
				}
			}/* Loop: All states and all incoming paths are valid */

		}/* Loop over all states */

		/* Make a decision on the bit transmitted 'Delay' bit periods before */
		if(k >= Delay-1)
		{
			/** Deciding on the transmitted bit based on detection delay **/
			/* Finding the minimum path metric at the current time */
			MinPathMetricIndex = 0;
			for(i=1;i<NumStates;i++){
				if(PathMetric[1][i] < PathMetric[1][MinPathMetricIndex])
					MinPathMetricIndex = i;
			}
			/* Detected output = first bit in Path[1][MinPathMetricIndex] */
			/* Path[1][MinPathMetricIndex] has 'Delay' number of bits */
			DetectedHardOutput[k-(Delay-1)] = (Path[1][MinPathMetricIndex] >> (Delay-1)) & 1;

			/* Detecting soft output */
			if(DetectedHardOutput[k-(Delay-1)] == 1)
				DetectedSoftOutput[k-(Delay-1)] = ProbWrongDet[1][MinPathMetricIndex][0];
			else
				DetectedSoftOutput[k-(Delay-1)] = 1 - ProbWrongDet[1][MinPathMetricIndex][0];
		}

		/* Exchange old and new Path and PathMetric values */
		for(i=0;i<NumStates;i++)
		{
			Path[0][i] = Path[1][i];
			Path[1][i] = 0;
			PathMetric[0][i] = PathMetric[1][i];
			PathMetric[1][i] = 0.0;
		}

		for(i=0;i<NumStates;i++)
		{
			for(j=0;j<Delay;j++)
			{
				ProbWrongDet[0][i][j] = ProbWrongDet[1][i][j];
				ProbWrongDet[1][i][j] = 0.0;
			}
		}

		/* Exchange old and new Valid State information */
		if(k < (PRImpResLength - 1))
		{
			VS[0] = (int*)realloc(VS[0],VSLengthCurrentTime*sizeof(int));
			for(i=0;i<VSLengthCurrentTime;i++)
			{
				VS[0][i] = VS[1][i];
				VS[1][i] = 0;
			}
			VSLengthPrevTime = VSLengthCurrentTime;
			VSLengthCurrentTime = 0;
		}

	}/* Loop over PaddedEqOutput */
	free(Path[0]);
	free(Path[1]);
	free(VS[0]);
	free(VS[1]);
	free(PathMetric[0]);
	free(PathMetric[1]);
	for(i=0;i<NumStates;i++)
	{
		free(ProbWrongDet[0][i]);
		free(ProbWrongDet[1][i]);
	}
	free(ProbWrongDet[0]);
	free(ProbWrongDet[1]);
}


/*-------------------------------------- CHANNEL WITH TRANSITION NOISE ---------------------------------------*/

void MediaNoiseFilter(double *OSPadEncodedBits, int SectorLength, double *OSChannelOutput, double *ReadChannelCoeff,
					   int NumChannelTaps, double *TimeIndex, double ND, double SigmaJitter, double SigmaPulseBroad,
					   int OSR, char *ChannelType)
{
	int i,j,k;
	int NumTrans = 0, TempNumTrans = 0;
	int *DiffOSPadEncodedBits;
	int FrontPadLength = 0, BackPadLength = 0;
	int *PaddedData = NULL;
	int CurrentDataTapLoc = 0;
	static long int idum3 = -200;
	static long int idum4 = -100;
	double *DeltaX, *DeltaND;
	double CurrentTapValue =0.0;

	/* Differentiate the bipolar input data. Array elements could be {-1,0,+1} */
	DiffOSPadEncodedBits = (int*)malloc(SectorLength*sizeof(int));
	DiffOSPadEncodedBits[0]= 0;
	for(i=1;i<SectorLength;i++)
	{
		/* 
		When differentiating, do not normalized by 2, in order to be consistent with 
		the output using impulse response of the channel
		*/
		DiffOSPadEncodedBits[i]= (OSPadEncodedBits[i] - OSPadEncodedBits[i-1]); 
	}

	/** Calculate the number of transitions that occur in the whole sequence **/
	DeltaX = (double*)malloc(1*sizeof(double));
	DeltaND = (double*)malloc(1*sizeof(double));
	for(i=0;i<SectorLength;i++)
	{/* Loops over input sequence */
		if(DiffOSPadEncodedBits[i] != 0)
		{
			NumTrans++;
			/* When a transition is detected, calculate the corresponding DeltaX and DeltaND */
			DeltaX = (double*)realloc(DeltaX, NumTrans*sizeof(double));
			DeltaND = (double*)realloc(DeltaND, NumTrans*sizeof(double));

			/* 
			E[DeltaT^2] = SigmaJitter % of T 
			E[(DeltaT/T)^2] = SigmaJitter/100
			Therefore, E[DeltaX^2] = SigmaJitter/100.

			-T/2 < DeltaT < T/2
			=> -1/2 < DeltaT/T < 1/2
			=> -1/2 < DeltaX < 1/2
			Similarly, for DeltaW and DeltaND.
			DeltaND = DeltaW/T.

			Therefore, we dont need to know T
			*/

			/* DeltaX PDF: Truncated Gaussian of Sigma = SigmaJitter/100 */
			/* -1/2 < DeltaX < 1/2 */
			DeltaX[NumTrans-1] = 0.5;
			while(fabs(DeltaX[NumTrans-1]) >= 0.5)
			{
				DeltaX[NumTrans-1] = (SigmaJitter/100)*GaussianRandomNumberGeneratorJitter(&idum3);
			}
			/* All time index are normalized by T */
			//DeltaX[NumTrans-1] = DeltaX[NumTrans-1]/T;

			DeltaND[NumTrans-1] = -1;
			/* DeltaND PDF: Single-Sided Gaussian */
			/* DeltaND >= 0 */
			while(DeltaND[NumTrans-1] < 0)
			{
				DeltaND[NumTrans-1] = (SigmaPulseBroad/100)*GaussianRandomNumberGeneratorPulse(&idum4);
			}
			/* All time index are normalized by T */
			//DeltaND[NumTrans-1] = DeltaND[NumTrans-1]/T;
		}
	}/* Loops over input sequence */

	/*** FIR Filtering ***/
	/** Padding Data for ease of FIR filtering operation **/
	/* h(0) is assumed to be at location h(floor(NumTaps/2)) */
	FrontPadLength = floor(NumChannelTaps/2.0);
	BackPadLength = NumChannelTaps-1;
	PaddedData = (int *)malloc((SectorLength + FrontPadLength + BackPadLength)*sizeof(int));

	/* PaddedData: < floor(NumTaps/2) zero's >< Data >< (NumTaps-1) zero's > */
	for(i=0;i<(SectorLength+FrontPadLength+BackPadLength);i++)
	{
		if(i<=FrontPadLength-1 || i >= FrontPadLength+SectorLength)
		{
			PaddedData[i] = 0;
		}else
		{
			PaddedData[i] = DiffOSPadEncodedBits[i-FrontPadLength];
		}

	}

	/*
	h: ...... h(-2) h(-1) h(0) h(1) h(2)......
	PaddedData: 0000....Data....000000
	First valid output: ...... + h(-2)*0 + h(-1)*0 + h(0)*Data(0) + h(1)*Data(1) +.....
	Last valid output: h(-floor(NumChannelTaps/2))*Data(DataLength-1) + h(-floor(NumChannelTaps/2)+1)*0 + .....
	*/
	NumTrans = 0;
	CurrentDataTapLoc = floor(NumChannelTaps/2.0);	/* Location of h(0) */
	for(i=0;i<(SectorLength+FrontPadLength);i++)
	{/* Loops through all input bits */

		OSChannelOutput[i] = 0.0;
		/* Checking if the current input (i.e corresponding to h(0) ) is a transition */
		if(PaddedData[i+NumChannelTaps/2] != 0)
		{
			NumTrans++;
			/* Calculate h(0)*CurrentBit */
			if(!strcmp(ChannelType,"perpendicular") || !strcmp(ChannelType,"Perpendicular"))
			{
				CurrentTapValue = (double) erf(2/ND * sqrt(log(2)) * (TimeIndex[CurrentDataTapLoc]+DeltaX[NumTrans-1]));
//				CurrentTapValue = CurrentTapValue/(double)OSR;
			}else if(!strcmp(ChannelType,"longitudinal") || !strcmp(ChannelType,"Longitudinal"))
			{
				CurrentTapValue = sqrt(2.0/PI)/(1+(4*pow(TimeIndex[CurrentDataTapLoc]+DeltaX[NumTrans-1],2)/pow(ND+DeltaND[NumTrans-1],2)));
//				CurrentTapValue = CurrentTapValue/(double)OSR;
			}
			OSChannelOutput[i] += PaddedData[i+NumChannelTaps/2]*CurrentTapValue;
		}

		/*
		For the current bit, calculate the filter output contributed by future inputs,
		i.e. because of h(1),h(2) etc.
		*/
		TempNumTrans = NumTrans;
		for(j=CurrentDataTapLoc+1;j<NumChannelTaps;j++)	/* h(1)to h(end) */
		{
			/* Need to calculate filter output only if the future input is non-zero */
			if(PaddedData[i+j] != 0)
			{
				TempNumTrans++;
				if(!strcmp(ChannelType,"perpendicular") || !strcmp(ChannelType,"Perpendicular"))
				{
					CurrentTapValue = (double) erf(2/ND * sqrt(log(2)) * (TimeIndex[j]+DeltaX[TempNumTrans-1]));
//					CurrentTapValue = CurrentTapValue/(double)OSR;
				}else if(!strcmp(ChannelType,"longitudinal") || !strcmp(ChannelType,"Longitudinal"))
				{
					CurrentTapValue = sqrt(2.0/PI)/(1+(4*pow(TimeIndex[j]+DeltaX[TempNumTrans-1],2)/pow(ND+DeltaND[TempNumTrans-1],2)));
//					CurrentTapValue = CurrentTapValue/(double)OSR;
				}
				OSChannelOutput[i] += PaddedData[i+j]*CurrentTapValue;
			}
		}

		/*
		For the current bit, calculate the filter output contributed by past inputs,
		i.e. because of h(-1),h(-2) etc.
		*/

		if(PaddedData[i+NumChannelTaps/2] != 0)
		{
			TempNumTrans = NumTrans;
		}else
		{
			TempNumTrans = NumTrans+1;
		}
		for(j=CurrentDataTapLoc-1;j>=0;j--)	/* h(-1) to h(start) */
		{
			/* Need to calculate filter output only if the past input is non-zero */
			if(PaddedData[i+j] != 0)
			{
				TempNumTrans--;
				if(!strcmp(ChannelType,"perpendicular") || !strcmp(ChannelType,"Perpendicular"))
				{
					CurrentTapValue = (double) erf(2/ND * sqrt(log(2)) * (TimeIndex[j]+DeltaX[TempNumTrans-1]));
//					CurrentTapValue = CurrentTapValue/(double)OSR;
				}else if(!strcmp(ChannelType,"longitudinal") || !strcmp(ChannelType,"Longitudinal"))
				{
					CurrentTapValue = sqrt(2.0/PI)/(1+(4*pow(TimeIndex[j]+DeltaX[TempNumTrans-1],2)/pow(ND+DeltaND[TempNumTrans-1],2)));
//					CurrentTapValue = CurrentTapValue/(double)OSR;
				}
				OSChannelOutput[i] += PaddedData[i+j]*CurrentTapValue;
			}
		}
	}/* Loops through all input bits */
	free(DiffOSPadEncodedBits);
	free(DeltaX);
	free(DeltaND);
	free(PaddedData);
}


/*------------------------------------- BISECTION -------------------------------------*/

#define JMAX 2000 			/* Set to maximum number of iterations */

/* Using bisection, find the root of a function known to lie between x1 and x2. The root,
/* returned as rtbis, will be refined until its accuracy is xacc. */
float Bisection(float x1, float x2, float xacc, double sigma_t,
			double T_Peak, double c, double d, double z0, double *Hc, double Hg, float g, float y)
{

int j;
double dx,f,fmid,xmid,rtb;
double Hc_x, Hh_x;		/* functions Hc(T(x)) and Hh(x) */

Hc_x = Hc[0]*(T_Peak*exp(-pow((z0-d),2)/(2*pow(sigma_t,2)))*exp(-pow((x1-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1];
Hh_x = (Hg/PI)*(atan((x1+g/2)/y) - atan((x1-g/2)/y));
f = Hc_x - Hh_x;

Hc_x = Hc[0]*(T_Peak*exp(-pow((z0-d),2)/(2*pow(sigma_t,2)))*exp(-pow((x2-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1];
Hh_x = (Hg/PI)*(atan((x2+g/2)/y) - atan((x2-g/2)/y));
fmid = Hc_x - Hh_x;


if (f*fmid >= 0.0){
	/* Try with x2 = max(0,c). Keep the x2 same */
	if(x2==0)
		x2 = c;
	else
		x2 = 0;
	Hc_x = Hc[0]*(T_Peak*exp(-pow((z0-d),2)/(2*pow(sigma_t,2)))*exp(-pow((x2-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1];
	Hh_x = (Hg/PI)*(atan((x2+g/2)/y) - atan((x2-g/2)/y));
	fmid = Hc_x - Hh_x; 
	/* Check again if there is any root in this new interval */
	if (f*fmid >= 0.0){
		fprintf(stderr,"Error: No solution exists for Hc(T(x)) = Hh(x)\n");
		exit(1);
	}
}

rtb = f < 0.0 ? (dx=x2-x1,x1) : (dx=x1-x2,x2); /* Orient the search so that f>0
												/* lies at x+dx */
for (j=1;j<=JMAX;j++) {
	xmid = rtb+(dx *= 0.5);
	Hc_x = Hc[0]*(T_Peak*exp(-pow((z0-d),2)/(2*pow(sigma_t,2)))*exp(-pow((xmid-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1];
	Hh_x = (Hg/PI)*(atan((xmid+g/2)/y) - atan((xmid-g/2)/y));
	fmid = Hc_x - Hh_x;
	if (fmid <= 0.0)
		rtb = xmid;
	if (fabs(dx) < xacc || fmid == 0.0){
		return rtb;
	}
}
printf("Maximum number of iterations run without converging \n");
exit(2);
return 0.0;			/* Never Reached! */
}


#define BISECTION_ACC 0.01	/* Accuracy of BISECTION methods */
#define ITERATION_ACC 0.1  /* Accuracy for x0 and a */

/*------------------------------------- HD -----------------------------------*/

/*
Calculates the demagnetizing field at a particular location, given transition centre and
parameter and other parameters. This function uses the formula given in Eq 2.25 
in Tim Rausch's PhD thesis.
*/
#define YSTART -500
#define YEND 500	/* SHOULD be -YSTART */
#define YSTEP 0.1

double Hd(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,
	      double x0, double a, double x)
{
	double y;
	double T_y, dMy, Hstep, P=0.0;

	/*** Calculating p(x) = -dM(x) @ Hstep(x) at x0, where @ => convolution ***/
	/* 
	p(x) = integral[ {-dM(y)*Hstep(x-y)}.dy) ]
	dMx : Eq 2.20 Pg 22, Tim's PhD thesis
	Hstep : Eq 2.26, Pg 24, Tim's PhD thesis
	*/
	for(y=YSTART ; y <= YEND ; y=y+YSTEP)
	{/* Loops over YSTART to YEND */

		/* Calculating dM at x */
		T_y = mp->T_Peak*exp(-pow((mp->z0-mp->d),2)/(2*pow(mp->sigma_t,2)))*exp(-pow((y-mp->c),2)/(2*pow(mp->sigma_t,2))) + 300;
		dMy = (2/PI)*-fabs(mp->Mr[0]*T_y + mp->Mr[1]) * (a/(a*a + (y-x0)*(y-x0)));
		dMy = dMy + (2/PI)*atan((y-x0)/a)*fabs(mp->Mr[0])*(T_y-300)*-(y-mp->c)/(pow(mp->sigma_t,2));
		dMy = dMy*YSTEP;	/* Oversampling by 1/XSTEP */

		/* Calculating Hstep at (x-x0) */
		if(x-y != 0)
		{
			Hstep = (1/PI)*atan(pp->t/(2*(x-y)));
		}
		else
		{
			/* 1/0 encountered */
			Hstep = 0.5;
		}

		/* Calculating P(x0) */
		P = P-dMy*Hstep;

	}/* Loops over YSTART to YEND */
	return P;
}


/*------------------------------------- HD GRADIENT -----------------------------------*/

/* Calculates the gradient of dHd/dx at a particular location */
double dHd(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,
	      double x0, double a, double x)
{
	double Hd1, Hd2;

	/* dHd(x) = (Hd(x+y) - Hd(x-y))/(2y) for some small y */
	Hd1 = Hd(mp,pp,rp,x0,a,x-YSTEP);
	Hd2 = Hd(mp,pp,rp,x0,a,x+YSTEP);

	return ((Hd2-Hd1)/(2*YSTEP));
}


/*-------------------------------- BISECTION TCENTRE --------------------------------*/

/* Using bisection, find the root of a function known to lie between x1 and x2. The root,
/* returned as rtbis, will be refined until its accuracy is xacc. */
#define JMAX 2000 			/* Set to maximum number of iterations */
#define NEG_LIMIT -500
#define POS_LIMIT 500

double BisectionTCentre(double x1, double x2, double xacc, struct Mag_Param *mp, struct Physical_Param *pp, 
						struct Reader_Param *rp, double a, double *TCentre, double *TParam, int NumTransition,
						int NumNLTSInfluencingTrans, int *TransBitLoc, double OversampledBitLength)
{
int i,j,temp=0;
double B;
double dx,f,fmid,xmid,rtb;
double Hc_x, Hh_x, Hd_x;		/* functions Hc(T(x)), Hh(x), Hd(x) */
double Hd_PrevTrans=0.0;		/* HDemag field of previous transitions */
double sigma_t = mp->sigma_t;
double T_Peak = mp->T_Peak;
double c = mp->c;
double z0 = mp->z0;
double Hg = -mp->Hg; /* NOTE the change in sign */
double Hc[2];
double g = pp->g;
double y = pp->y;

Hc[0] = mp->Hc[0];
Hc[1] = mp->Hc[1]; 

/* Solving for x0 such that Hd(x0)+ Hh(x0) = Hc(x0) */
Hc_x = -fabs(Hc[0]*(T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x1-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1]);
Hh_x = (Hg/PI)*(atan((x1+g/2)/y) - atan((x1-g/2)/y));	/* Will take the sign of Hg */
Hd_x = Hd(mp,pp,rp,x1,a,x1);

/* Calculate HDemag of previous transitions (i.e include NLTS effects) */
if(NumTransition > 1)
{
	temp = 0;
	Hd_PrevTrans = 0.0;
	for(i=NumTransition-1 ; i>=(NumTransition - 1 >= NumNLTSInfluencingTrans ? NumTransition-NumNLTSInfluencingTrans : 1) ; i--)
	{
		/* Calculate the transition centre of the transition i, w.r.t the gap centre of the current transition */
		B = -((TransBitLoc[NumTransition] - TransBitLoc[i])*OversampledBitLength + fabs(TCentre[i-1]));
		Hd_PrevTrans += pow(-1,temp) * -Hd(mp,pp,rp,B,TParam[i-1],x1);
		temp++;
	}
}

f = -(Hc_x - Hh_x - Hd_x - Hd_PrevTrans);

Hc_x = -fabs(Hc[0]*(T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x2-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1]);
Hh_x = (Hg/PI)*(atan((x2+g/2)/y) - atan((x2-g/2)/y));
Hd_x = Hd(mp,pp,rp,x2,a,x2);
/* Calculate HDemag of previous transitions (i.e include NLTS effects) */
if(NumTransition > 1)
{
	temp = 0;
	Hd_PrevTrans = 0.0;
	for(i=NumTransition-1 ; i>=(NumTransition - 1 >= NumNLTSInfluencingTrans ? NumTransition-NumNLTSInfluencingTrans : 1) ; i--)
	{
		/* Calculate the transition centre of the transition i, w.r.t the gap centre of the current transition */
		B = -((TransBitLoc[NumTransition] - TransBitLoc[i])*OversampledBitLength + fabs(TCentre[i-1]));
		Hd_PrevTrans += pow(-1,temp) * -Hd(mp,pp,rp,B,TParam[i-1],x2);
		temp++;
	}
}
fmid = -(Hc_x - Hh_x - Hd_x - Hd_PrevTrans);

if (f*fmid >= 0.0)
{
	/* Try with x2 = max(0,c). Keep the x2 same */
	if(x2==0)
		x2 = c;
	else
		x2 = 0;
	Hc_x = -fabs(Hc[0]*(T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x2-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1]);
	Hh_x = (Hg/PI)*(atan((x2+g/2)/y) - atan((x2-g/2)/y));
	Hd_x = Hd(mp,pp,rp,x2,a,x2);
	/* Calculate HDemag of previous transitions (i.e include NLTS effects) */
	if(NumTransition > 1)
	{
		temp = 0;
		Hd_PrevTrans = 0.0;
		for(i=NumTransition-1 ; i>=(NumTransition - 1 >= NumNLTSInfluencingTrans ? NumTransition-NumNLTSInfluencingTrans : 1) ; i--)
		{
			/* Calculate the transition centre of the transition i, w.r.t the gap centre of the current transition */
			B = -((TransBitLoc[NumTransition] - TransBitLoc[i])*OversampledBitLength + fabs(TCentre[i-1]));
			Hd_PrevTrans += pow(-1,temp) * -Hd(mp,pp,rp,B,TParam[i-1],x2);
			temp++;
		}
	}
	fmid = -(Hc_x - Hh_x - Hd_x - Hd_PrevTrans); 
	/* Check again if there is any root in this new interval */
	if (f*fmid >= 0.0)
	{
		fprintf(stderr,"Error: No solution exists for Hc(T(x)) = Hh(x)\n");
		exit(-1);
	}
}

rtb = f < 0.0 ? (dx=x2-x1,x1) : (dx=x1-x2,x2); /* Orient the search so that f>0 
												/* lies at x+dx */
for (j=1;j<=JMAX;j++) {
	xmid = rtb+(dx *= 0.5);
	Hc_x = -fabs(Hc[0]*(T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((xmid-c),2)/(2*pow(sigma_t,2))) + 300) + Hc[1]);
	Hh_x = (Hg/PI)*(atan((xmid+g/2)/y) - atan((xmid-g/2)/y));
	Hd_x = Hd(mp,pp,rp,xmid,a,xmid);
	/* Calculate HDemag of previous transitions (i.e include NLTS effects) */
	if(NumTransition > 1)
	{
		temp = 0;
		Hd_PrevTrans = 0.0;
		for(i=NumTransition-1 ; i>=(NumTransition - 1 >= NumNLTSInfluencingTrans ? NumTransition-NumNLTSInfluencingTrans : 1) ; i--)
		{
			/* Calculate the transition centre of the transition i, w.r.t the gap centre of the current transition */
			B = -((TransBitLoc[NumTransition] - TransBitLoc[i])*OversampledBitLength + fabs(TCentre[i-1]));
			Hd_PrevTrans += pow(-1,temp) * -Hd(mp,pp,rp,B,TParam[i-1],xmid);
			temp++;
		}
	}
	fmid = -(Hc_x - Hh_x - Hd_x - Hd_PrevTrans);						
	if (fmid <= 0.0) 
		rtb = xmid;
	if (fabs(dx) < xacc || fmid == 0.0){
		return rtb;
	}
}
printf("Maximum number of iterations run without converging \n");
exit(2);
return 0.0;			/* Never Reached */
}


/*-------------------------------- BISECTION TPARAM --------------------------------*/


/* Using bisection, find the root of a function known to lie between x1 and x2. The root,
/* returned as rtbis, will be refined until its accuracy is xacc. */
#define JMAX 2000 			/* Set to maximum number of iterations */
#define NEG_LIMIT -500
#define POS_LIMIT 500

double BisectionTParam(double x1, double x2, double xacc, struct Mag_Param *mp, struct Physical_Param *pp, 
						struct Reader_Param *rp, double x0)
{
int j;
double dx,f,fmid,xmid,rtb;
double dM_dx, dM_dH, dHh_dx, dHd_dx, dHc_dT, dT_dx;	/* functions dM/dx, dM/dH, d(Hd)/dx, d(Hc)/dT, dT/dx */
double Mr0,Hc0,T0,Q;		/* Mr(T0), Hc(T0), T0 and Q */
double sigma_t = mp->sigma_t;
double T_Peak = mp->T_Peak;
double c = mp->c;
double z0 = mp->z0;
double Hg = -mp->Hg;	/* NOTE the -ve sign */
double Hc[2], Mr[2], S[2];
double g = pp->g;
double y = pp->y;

Hc[0] = mp->Hc[0];
Hc[1] = mp->Hc[1]; 
Mr[0] = mp->Mr[0];
Mr[1] = mp->Mr[1];
S[0] = mp->S[0];
S[1] = mp->S[1];

/* Solving for 'a' such that Thermal-Williams slope eqn (Eqn 2.17 in Tim Rausch Thesis) is satisfied @ x0,T0 */
T0 = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2))) + 300;
Mr0 = -fabs(Mr[0]*T0 + Mr[1]);
Hc0 = -fabs(Hc[0]*T0 + Hc[1]);

dM_dH = fabs(Mr0/(Hc0*(1-S[0]*T0-S[1])));
dHh_dx = Hg/(PI*y) * (1/(1+pow((x0+g/2)/y,2)) - 1/(1+pow((x0-g/2)/y,2)));	/* Will take the sign of Hg */
dHc_dT = fabs(Hc[0]);
dT_dx = (T0-300)*-(x0-c)/(sigma_t*sigma_t);

/* Only these quantities are dependent on tansition parameter */
if(x1!=0)
	dM_dx = 2*Mr0/(PI*x1);
else
	dM_dx = -1e+50;	/* -ve because dM_dx is -ve */

/* d(Hd)/dx is evaluated at x0 with transition parameter x1 */
dHd_dx = dHd(mp,pp,rp,x0,x1,x0);
/* f(x1) */
f = dM_dx - dM_dH*(dHh_dx + dHd_dx - dHc_dT*dT_dx);

/* f(x2) */
/* Recalculate only those quantities that are dependent on transition parameter */
if(x2!=0)
	dM_dx = 2*Mr0/(PI*x2);
else
	dM_dx = -1e+50;

dHd_dx = dHd(mp,pp,rp,x0,x2,x0);
fmid = dM_dx - dM_dH*(dHh_dx + dHd_dx - dHc_dT*dT_dx);

if (f*fmid >= 0.0){
	/* Try with x2 = max(0,c). Keep the x2 same */
	if(x2==0)
		x2 = c;
	else
		x2 = 0;

	if(x2!=0)
		dM_dx = 2*Mr0/(PI*x2);
	else
		dM_dx = -1e+50;

	dHd_dx = dHd(mp,pp,rp,x0,x2,x0);
	fmid = dM_dx - dM_dH*(dHh_dx + dHd_dx - dHc_dT*dT_dx);
	/* Check again if there is any root in this new interval */
	if (f*fmid >= 0.0){
		fprintf(stderr,"Error: No solution exists for Thermal-Williams Slope Equation\n");
		exit(1);
	}
}

rtb = f < 0.0 ? (dx=x2-x1,x1) : (dx=x1-x2,x2); /* Orient the search so that f>0 
												/* lies at x+dx */
for (j=1;j<=JMAX;j++) {
	xmid = rtb+(dx *= 0.5);
	if(xmid!=0)
		dM_dx = 2*Mr0/(PI*xmid);
	else
		dM_dx = -1e+50;

	dHd_dx = dHd(mp,pp,rp,x0,xmid,x0);
	fmid = dM_dx - dM_dH*(dHh_dx + dHd_dx - dHc_dT*dT_dx);
	if (fmid <= 0.0) 
		rtb = xmid;
	if (fabs(dx) < xacc || fmid == 0.0){
		return rtb;
	}
}
printf("Maximum number of iterations run without converging \n");
exit(2);
return 0.0;			/* Never Reached */
}


/*---------------------------- TRANSITION LARGE SPOT APPROX -------------------------*/

void TransitionLargeSpot(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,
				double *TCentre, double *TParam)
{

double sigma_t = mp->sigma_t;		/* Sigma of T-Profile in nm */
double T_Peak = mp->T_Peak;			/* Peak Tempearture in centigrade */
double c = mp->c;					/* Shift in peak temperature w.r.t gap centreline in nm*/
double z0 = mp->z0;					/* T_calc = T_peak * exp(-z0/(2*sigma_t^2)) */
double *Hc = &(mp->Hc);				/* Dependence of coercivity on Temperature in A/m. Hc = aT + b => {a,b} */
double *Mr = &(mp->Mr);				/* Dependence of remnant magnetization on Temperature in A/m. Mr = aT + b => {a,b} */
double *S = &(mp->S);				/* Dependence of coercive squareness on Temperature. S = aT + b => {a,b} */
double Hg = mp->Hg;					/* Deep gap field in A/m*/

/* Physical Parameters */
float g = pp->g;					/* Write/Head gap in nm*/
float d = pp->d;					/* Medium spacing(in nm) or the space between the bottom of pole and the surface of the medium */
float t = pp->t;					/* Medium thickness in nm */
float y = pp->y;					/* Distance from bottom of pole to the centre of the medium in nm */

/* Transition Parameters */
double x0; //TCentre;
double a; //TParam;

/* Other variables */
double temp1, temp2, T0, Q, beta, dT, Hc_x0, Mr_x0, dHc;

/* Calculation of transition centre w.r.t the gap centreline (i.e. x = 0) in the downtrack  location.
/* Ha(Xo) = Hh(Xo) + Hd(Xo), where
/* Ha: Applied field, Hh: Head field, Hd: Demagnetizing field, Xo: transition centre
/* By simplification explained in Pg 33-36 of Tim Rausch's PhD thesis,
/* Hd(Xo) can be neglected at Xo. Therefore,
/* Ha(Xo) = Hc(Xo) = Hc(T(Xo)) = Hh(Xo)
/* This equation is solved for Xo using the Bisection method */

/* Decide the interval [x1 x2] in which the root Xo lies
/* This is taken to be [max(0,c) max(0,c)+5*Sigma_t]
/* 5*sigma_t is arbitrary assuming that the root will lie before that. This is a reasonable estimate because,
/* 5*Sigma_t if far away from the maxima/minima of the coercivity (near which the root is likely to be found) */
/* The root in this interval will also be the root (in case of multiple roots) that is farthest to the 
/* right of the gap centreline */

float x1,x2;

/* x2 = min(0,c) */
if (c > 0){
	x2 = 0.0;
}else{
	x2 = (float)c;
}

x1 = x2 - (float)5*sigma_t;

/* Solve numerically using bisection method
/* This method is adpated from the function given in 
/* pg. 353 of the book 'Numerical Recipes in C' */

x0 = Bisection(x1, x2, BISECTION_ACC, sigma_t, T_Peak, c, mp->d, z0, Hc, Hg, g, y);
*TCentre = x0;

/* Evaluation of transition parameter
/* Using the formula given in Pg 37 of Tim Rausch's PhD thesis 
/* Transition considered here is from +Mr to -Mr 
/* Therefore, Hc: -ve, dHc/dt: +ve, Mr: +ve */
T0 = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2))) + 300;
dT = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2)))*(-x0+c)/(pow(sigma_t,2));
Hc_x0 = -fabs(Hc[0]*T0 + Hc[1]); /* -ve because transition is from +Mr to -Mr */
dHc = fabs(Hc[0]);				 /* +ve because transition is from +Mr to -Mr */
Mr_x0 = -fabs(Mr[0]*T0 + Mr[1]);  /* +ve because transition is from +Mr to -Mr */
Hg = -Hg;						 /* Applied field is -ve, therefore Hg is -ve too */

Q = pow(sin(PI*Hc_x0/Hg),2)*2*x0*Hg/(PI*g*Hc_x0); 
beta = 1 - dHc*dT*y/(fabs(Hc_x0)*Q);
temp1 = -y*(1-(S[0]*T0 + S[1]))/(PI*Q*beta);
temp2 = Mr_x0*t*y/(PI*Q*fabs(Hc_x0)*beta);
a =  temp1 - t/4 + sqrt(pow((temp1-t/4),2) + temp2 + t*temp1);
*TParam = a;

/* Checking */
/* This value should be zero if correct TParam is calculated */
temp1 = 2*Mr_x0/(PI*a) - fabs(Mr_x0/(Hc_x0*(1-(S[0]*T0 + S[1]))))*(-Q*Hc_x0/y - Mr_x0*t/(PI*a*(a+t/2)) - dHc*dT);
//printf("diff = %g\n",temp1);

return;
}


/*---------------------------------- VALIDATE ANSWER --------------------------------*/

double ValidateAnswer(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,
					double x0, double a)
{
double f;
double dM_dx, dM_dH, dHh_dx, dHd_dx, dHc_dT, dT_dx;	/* functions dM/dx, dM/dH, d(Hd)/dx, d(Hc)/dT, dT/dx */
double Mr0,Hc0,T0,Q;		/* Mr(T0), Hc(T0), T0 and Q */
double sigma_t = mp->sigma_t;
double T_Peak = mp->T_Peak;
double c = mp->c;
double z0 = mp->z0;
double Hg = -mp->Hg;	/* NOTE the -ve sign */
double Hc[2], Mr[2], S[2];
double g = pp->g;
double y = pp->y;

Hc[0] = mp->Hc[0];
Hc[1] = mp->Hc[1]; 
Mr[0] = mp->Mr[0];
Mr[1] = mp->Mr[1];
S[0] = mp->S[0];
S[1] = mp->S[1];

T0 = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2))) + 300;
Mr0 = -fabs(Mr[0]*T0 + Mr[1]);
Hc0 = -fabs(Hc[0]*T0 + Hc[1]);

dM_dH = fabs(Mr0/(Hc0*(1-S[0]*T0-S[1])));
dHh_dx = Hg/(PI*y) * (1/(1+pow((x0+g/2)/y,2)) - 1/(1+pow((x0-g/2)/y,2)));	/* Will take the sign of Hg */
dHc_dT = fabs(Hc[0]);
dT_dx = (T0-300)*-(x0-c)/(sigma_t*sigma_t);

dM_dx = 2*Mr0/(PI*a);

/* d(Hd)/dx is evaluated at x0 with transition parameter a */
dHd_dx = dHd(mp,pp,rp,x0,a,x0);
/* Thermal Williams Comstock slope equation */
/* This should be 0 if x0 and a are correct */
f = dM_dx - dM_dH*(dHh_dx + dHd_dx - dHc_dT*dT_dx);

return f;
}


/*---------------------------------------- TRANSITION --------------------------------*/


void Transition(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,
				double *TCentre, double *TParam, int NumTransition, int NumNLTSInfluencingTrans,
				int *TransBitLoc, double OversampledBitLength)
{

int MaxNumIter = 20;
int NumIter=0;
int i;
double diffx0=1.0, diffa=1.0;
double temp;
double x0[2],a[2];	/* Transition Parameters */

/* Calculation of transition centre w.r.t the gap centreline (i.e. x = 0) in the downtrack  location.
/* Ha(Xo) = Hh(Xo) + Hd(Xo), where
/* Ha: Applied field, Hh: Head field, Hd: Demagnetizing field, Xo: transition centre
/* This equation is solved for Xo using the Bisection method */

/* Decide the interval [x1 x2] in which the root Xo lies
/* This is taken to be [min(0,c)-5*Sigma_t min(0,c)]
/* 5*sigma_t is arbitrary assuming that the root will lie after that. This is a reasonable estimate because,
/* 5*Sigma_t if far away from the maxima/minima of the coercivity (near which the root is likely to be found) */
/* The root in this interval will also be the root (in case of multiple roots) that is farthest to the 
/* left of the gap centreline (the medium is moving from right to left) */

double x1,x2;
FILE *fp;

/* TEST */
double T0, Mr0;
FILE *fpTest;

/* x2 = min(0,c) */
if (mp->c > 0){
	x2 = 0.0;
}else{
	x2 = mp->c;
}

x1 = x2 - 5*(mp->sigma_t);

x0[0] = 0; /* Trans centre at the previous iteration */
x0[1] = 0; /* Trans centre at the current iteration */
a[0] = 0; /* Trans parameter at the previous iteration */
a[1] = 0; /* Trans parameter at the current iteration */

/* 
Making an initial transition with Large Spot approximation in order to determine
initial x0 and a values.
*/

TransitionLargeSpot(mp,pp,rp,&(x0[0]),&(a[0]));
/*
fp = fopen("TransitionFileLargeSpot.dat","a");
fprintf(fp,"%g %g %g\n",mp->z0,a[0],x0[0]);
fclose(fp);
*/

while((diffx0 > ITERATION_ACC) || (diffa > ITERATION_ACC))
{
	/* 
	Calculate transition centre using the transition centre 
	and parameter values of previous iteration.
	*/
	x1 = x0[0]-30;
	x2 = (mp->c < 0) ? mp->c:0;

	x0[1] = BisectionTCentre(x1,x2,BISECTION_ACC,mp,pp,rp,a[0],TCentre,TParam,
							NumTransition,NumNLTSInfluencingTrans,TransBitLoc,OversampledBitLength);
	diffx0 = fabs(x0[1]-x0[0]);
	
	/* 
	Calculate transition parameter using transition centre value 
	of the current iteration.
	*/
	x1 = (a[0]-20 < 2) ? 2:a[0]-20;
	x2 = a[0]+20;

	a[1] = BisectionTParam(x1, x2, BISECTION_ACC, mp, pp, rp, x0[1]);
	diffa = fabs(a[1]-a[0]);

	/* Store current iteration value as previous */
	x0[0] = x0[1];
	x0[1] = 0;
	a[0] = a[1];
	a[1] = 0;

	NumIter++;
	if(NumIter >= MaxNumIter)
	{
		printf("\nWARNING: Transition() -> Maximum Number of iterations reached. Results may not be accurate.\n");
		break;
	}
	temp = ValidateAnswer(mp,pp,rp,x0[0],a[0]);
	if(fabs(temp) < 10)
		break;
}

TCentre[NumTransition-1] = x0[0];
TParam[NumTransition-1] = a[0];

temp = ValidateAnswer(mp,pp,rp,x0[0],a[0]);

if(fabs(temp) > 200)
{
	printf("\nWARNING: Transition() -> Transition Centre and Parameter may not be correct. Error in slope equation is %g\n",temp);
}


return;
}


/*------------------------------- MICROTRACK ---------------------------------*/

#define ACC 1e-05	/* Accuracy of the Bisection method, which is used to determine the transition centre */
#define PI 3.1415926
#define DEBUG_MICROTRACK 0

void Microtrack(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,int N,
				double *x, double *v, double *PW, int length, int *PeakAmpPositionIndex,int NumTransition, double **TCentre, 
				double **TParam, double *k, double *rho, int *TransBitLoc,int NumNLTSInfluencingTrans, double OversampledBitLength,
				int OSR, int NumSectors, int TemperatureVariation, double SigmaTempVariation, double PeakTempTruncValue, 
				int WriteHdCrossTrackMov, double MaxWriteHdCrTrMov, double MeanWriteHdCrTrMov, double SigmaJitter, int HMD, 
				double SigmaHMDVariation, int NumARModelCoeff, double *ARModelCoeff, double *DeltaD, double *NormalizationFactor, 
				int TransToCalcTransParam, int TemperatureModulatedFlag)
{
int i,j,G;
double diffx;	/* Distance between some downtrack location and the transition centre x0 */
double delz;	/* Width of one microtrack */
double rs;		/* Reader sensitivity */
double *M;		/* Magnetization along the downtrack direction */

double sigma_t = mp->sigma_t;		/* Sigma of thermal profile */
double sigma_r = rp->sigma_r;		/* Sigma of reader sensitivity */
double T_Peak = mp->T_Peak;			/* Peak Tempearture in centigrade */
double c = mp->c;					/* Shift in peak temperature w.r.t gap centreline in nm*/
double *Hc = &(mp->Hc);				/* Dependence of coercivity on Temperature in A/m. Hc = aT + b => {a,b} */
double *Mr = &(mp->Mr);				/* Dependence of remnant magnetization on Temperature in A/m. Mr = aT + b => {a,b} */
double *S = &(mp->S);				/* Dependence of coercive squareness on Temperature. S = aT + b => {a,b} */
double Hg = mp->Hg;					/* Deep gap field in A/m */
double g = pp->g;					/* Write/Head gap in nm */
double d = pp->d;					/* Medium spacing(in nm) or the space between the bottom of pole and the surface of the medium */
double t = pp->t;					/* Medium thickness in nm */
double y = pp->y;					/* Distance from bottom of pole to the centre of the medium in nm */
double tr = rp->tr;
double gr = rp->gr;

double temp, min, max, z0;
double sigma_eq;
int index_max, left_index_half_max, right_index_half_max;
static int idum_temp = -600;
static int idum_WriteHdCrTr = -900;

double x0;	/* Transition Centre */
double a; 	/* Transition Parameter */

/* NLTS Parameters */
double delta;		/* NLTS Shift */
double B;			/* Distance between two transitions in a microtrack */

/* Other variables */
static long int idum5 = -160;
static long int idum6 = -280;
double DeltaX = 0.0;
double b;		/* Jitter */
double temp1, temp2, T0, Q, beta, dT, Hc_x0, Mr_x0, dHc;
float x1,x2;

FILE *fpDebug;
char DebugFileName[] = "DebugFile_Microtrack.dat";

delz = pp->wt/N;		/* Width of a microtrack */

/* Change Peak Temperature for the current transition, if temperature variation effect is ON */
if(TransToCalcTransParam == 0 && TemperatureVariation == 1)
{
	mp->T_Peak = 0;
	while(fabs(mp->T_Peak - mp->Orig_T_Peak) > PeakTempTruncValue)
	{
		/* Change peak temperature according to a truncated gaussian */
		mp->T_Peak = GaussianRandomNumberGeneratorTemp(&idum_temp)*(SigmaTempVariation/100 * mp->Orig_T_Peak) + mp->Orig_T_Peak;
	}
	T_Peak = mp->T_Peak;
}

/* 
Change Cross track location of peak temperature for the current transition, 
if 'write head cross track movement' effect is ON.
*/
if(TransToCalcTransParam == 0 && WriteHdCrossTrackMov == 1)
{
	mp->d = 10000; /* Some large number */
	while(fabs(mp->d - mp->Orig_d) > MaxWriteHdCrTrMov/100 * pp->wt)
	{
		/* Change cross track location of peak temperature according to a double exp dist */
		mp->d = mp->Orig_d - (MeanWriteHdCrTrMov/100 * pp->wt) * log(UniformRandomNumberGeneratorWrHdCrTrMov(&idum_WriteHdCrTr));
	}
}

/*
If the HMD variation effect is ON, then change the HMD distance (pp->d) and the associated
change in Sigma of Temp and and Peak Temp.
*/
if(TransToCalcTransParam == 0 && HMD == 1)
{
	
	/* TEMP CODE HARD CODED FOR AR ORDER 1 */
	b = sqrt(1 + ARModelCoeff[0]*ARModelCoeff[0] - 2*ARModelCoeff[0]);

	DeltaD[NumTransition-1]= b*SigmaHMDVariation*GaussianRandomNumberGeneratorHMD(&idum6);
	for(i=1;i<=(NumTransition >= NumARModelCoeff+1 ? NumARModelCoeff:NumTransition-1);i++)
	{
		DeltaD[NumTransition-1] = DeltaD[NumTransition-1] + ARModelCoeff[i-1]*DeltaD[NumTransition-1-i];
	}
	d = pp->d + DeltaD[NumTransition-1];
	sigma_t = mp->sigma_t * sqrt(1 + pow(DeltaD[NumTransition-1],2)/pow(pp->d,2));
}


/* Calculate jitter for the current transition if the variable 'SigmaJitter' is non-zero */
/* Do not add jitter when equalizer is adapting */
if(TransToCalcTransParam == 0 && SigmaJitter != 0.0 && NumSectors >= 0)
{
	DeltaX = 0.5;
	while(fabs(DeltaX) >= 0.5)
		DeltaX = (SigmaJitter/100)*GaussianRandomNumberGeneratorJitter(&idum5);
	DeltaX = DeltaX * OSR * OversampledBitLength;	/* DeltaX = shift in transition centre in nm */
}

/* Calculate h(ai,ti) for track i, where i = 1,2,3...,N and sum h(ai,ti) over all tracks */
for(i=1 ; i<= N ; i++){

	/* Determine z0 */
	mp->z0 = pp->wt/2 - (delz*(i-1) + delz/2);
	z0 = mp->z0;

	if(LARGE_SPOT)
	{ /* Large Spot */

		/* Calculation of transition centre w.r.t the gap centreline (i.e. x = 0) in the downtrack  location.
		/* Ha(Xo) = Hh(Xo) + Hd(Xo), where
		/* Ha: Applied field, Hh: Head field, Hd: Demagnetizing field, Xo: transition centre
		/* By simplification explained in Pg 33-36 of Tim Rausch's PhD thesis,
		/* Hd(Xo) can be neglected at Xo. Therefore,
		/* Ha(Xo) = Hc(Xo) = Hc(T(Xo)) = Hh(Xo)
		/* This equation is solved for Xo using the numerical newton-raphson method */

		/* Decide the interval [x1 x2] in which the root Xo lies
		/* This is taken to be [max(0,c) max(0,c)+5*Sigma_t]
		/* 5*sigma_t is arbitrary assuming that the root will lie before that. This is a reasonable estimate because,
		/* 5*Sigma_t if far away from the maxima/minima of the coercivity (near which the root is likely to be found) */
		/* The root in this interval will also be the root (in case of multiple roots) that is farthest to the
		/* right of the gap centreline */

		/* x2 = min(0,c) */
		if (c > 0)
		{
			x2 = 0.0;
		}else
		{
			x2 = (float)c;
		}

		x1 = x2 - (float)5*sigma_t;

		/* Solve numerically using bisection method
		/* This method is adpated from the function given in
		/* pg. 353 of the book 'Numerical Recipes in C' */
		if(NumTransition == 1)		/* Transition is not affected by NLTS */
		{
			if(TemperatureModulatedFlag == -1)	/* That is, the corresponding transition is temperature modulated */
			{
				/* If the first transition is modulated, then we need to calculate the transition centre with the original peak
				temperature and store it in TCentre[i-1][0] and calculate the trans centre with modulated peak temp and store it in 
				x0 (before reverting back to TCentre[i-1][0] while calculating readback signal), for calculating the correct trans param.
				*/
				T_Peak = mp->Orig_T_Peak;
				TCentre[i-1][0] = Bisection(x1, x2, ACC, sigma_t, T_Peak, c, mp->d, z0, Hc, Hg, g, y);
				T_Peak = mp->T_Peak;
			}
			else
			{
				/* For every sector, TCentre for the first transition is calculated. It is a variable which is malloced, initialized
				   and freed for every sector (for uncoded). */
				TCentre[i-1][0] = Bisection(x1, x2, ACC, sigma_t, T_Peak, c, mp->d, z0, Hc, Hg, g, y);
				x0 = TCentre[i-1][0]; /* IMP: This x0 is w.r.t the gap centre and NOT w.r.t position of peak readback amplitude */
			}
		}else
		{
			/* Correct location of x0 (i.e. unaffected by NLTS)
			 * for all tracks is stored in TCentre[0,1,...,N-1][0] 
			 */
			 delta = 0.0;
			 temp = 1;
			 TCentre[i-1] = (double*)realloc(TCentre[i-1],NumTransition*sizeof(double));
			 TCentre[i-1][NumTransition-1] = TCentre[i-1][0];	/* Original Tcentre without NLTS */
			 TParam[i-1] = (double*)realloc(TParam[i-1],NumTransition*sizeof(double));
			 TParam[i-1][NumTransition-1] = 0;

			 if(TemperatureVariation == 1 || WriteHdCrossTrackMov == 1)
			 {
				 /* Change in temperature (peak or alignment) requires recalculation of transition centre */
				 TCentre[i-1][NumTransition-1] = Bisection(x1, x2, ACC, sigma_t, T_Peak, c, mp->d, z0, Hc, Hg, g, y);
				 x0 = TCentre[i-1][NumTransition-1]; /* IMP: This x0 is w.r.t the gap centre and NOT peak readback amplitude */
			 }
				
			 /* Other transitions are affected by NLTS. So calc new location of x0 */
			 /* Loops over all past transitions that contributes to NLTS */
			 for(j = NumTransition-1 ; j >= (NumTransition - 1 >= NumNLTSInfluencingTrans ? NumTransition-NumNLTSInfluencingTrans : 1) ; j--)
			 {
				 B = (TransBitLoc[NumTransition] - TransBitLoc[j])*OversampledBitLength + TCentre[i-1][0] - TCentre[i-1][j-1] ; /* Tcentre[i-1][0] is the correct Trans Centre location of the new transition */
				 delta = pow(-1,temp)*k[i-1]/pow(B,rho[i-1]);	/* NLTS because of jth transition */

				 delta = delta * (mp->Mr[0]*(T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((-(B+fabs(TCentre[i-1][0]))-c),2)/(2*pow(sigma_t,2))) + 300) + mp->Mr[1])/(mp->Mr[0]*300 + mp->Mr[1]);	/* Delta = Delta * Mr(Temp of prev trans)/Mr(at room temperature) */

				 TCentre[i-1][NumTransition-1] += delta;
				 temp++;
			 }
			 x0 = TCentre[i-1][NumTransition-1]; /* IMP: This x0 is w.r.t the gap centre and NOT peak readback amplitude */
		}
		
		/* If temperature is modulated for the current transition, then its transition center is re-calculated,
		for the purpose of calculation of the transition parameter. Its center is not same as one stored in TCentre[i-1][0].
		BUT, while calculating the readback signal, transition centre(x0) is assumed to be TCentre[i-1][0], since in practice,
		even when the temperature is modulated, we assume that the location of write-head is adjusted to take care of this change
		in x0 */
		if(TemperatureModulatedFlag == -1)	/* That is, the corresponding transition is temperature modulated */
		{
			/* If the first transition is modulated, then we need to calculate the transition centre with the original peak
			temperature and store it in TCentre[i-1][0] and calculate the trans centre with modulated peak temp and store it in 
			x0 (before reverting back to TCentre[i-1][0] while calculating readback signal), for calculating the correct trans param.
			*/
			x0 = Bisection(x1, x2, ACC, sigma_t, T_Peak, c, mp->d, z0, Hc, Hg, g, y); /* IMP: This x0 is w.r.t the gap centre and NOT w.r.t position of peak readback amplitude */
			if(NumTransition==1 && NumSectors==0 && i==1)
				printf("WARNING: Microtrack() --> This function does not support temperature modulation in the presence of any noise other than AWGN and Position Jitter\n");
		}
		
		/* Evaluation of transition parameter.
		/* Using the formula given in Pg 37 of Tim Rausch's PhD thesis
		/* Transition considered here is from +Mr to -Mr (medium moving in the -x direction)
		/* Therefore, Hc: -ve, dHc/dt: +ve, Mr: -ve */

		T0 = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2))) + 300;
		dT = T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(sigma_t,2)))*exp(-pow((x0-c),2)/(2*pow(sigma_t,2)))*(-x0+c)/(pow(sigma_t,2));
		Hc_x0 = -fabs(Hc[0]*T0 + Hc[1]); /* -ve because transition is from +Mr to -Mr */
		dHc = fabs(Hc[0]);				 /* +ve because transition is from +Mr to -Mr */
		Mr_x0 = -fabs(Mr[0]*T0 + Mr[1]); /* -ve because transition is from +Mr to -Mr */
		//Hg = -Hg;						 /* Applied field is -ve */
		Q = pow(sin(PI*Hc_x0/(-Hg)),2)*2*x0*(-Hg)/(PI*g*Hc_x0);
		beta = 1 - dHc*dT*y/(fabs(Hc_x0)*Q);
		temp1 = -y*(1-(S[0]*T0 + S[1]))/(PI*Q*beta);
		temp2 = Mr_x0*t*y/(PI*Q*fabs(Hc_x0)*beta);
		a =  temp1 - t/4 + sqrt(pow((temp1-t/4),2) + temp2 + t*temp1);
		/* Checking for correct calculation of 'a' */
		/* This value should be zero if 'a' is calculated correctly */
		temp1 = 2*Mr_x0/(PI*a) - fabs(Mr_x0/(Hc_x0*(1-(S[0]*T0 + S[1]))))*(-Q*Hc_x0/y - Mr_x0*t/(PI*a*(a+t/2)) - dHc*dT);
		//printf("diff = %g\n",temp1);
		TParam[i-1][NumTransition-1] = a;

	}/* Large spot */
	else if(MEDIUM_SPOT)
	{
		TCentre[i-1] = (double*)realloc(TCentre[i-1],NumTransition*sizeof(double));
		TCentre[i-1][NumTransition-1] = TCentre[i-1][0];	/* Original Tcentre without NLTS */
		TParam[i-1] = (double*)realloc(TParam[i-1],NumTransition*sizeof(double));
		TParam[i-1][NumTransition-1] = 0;

		Transition(mp,pp,rp,TCentre[i-1],TParam[i-1],NumTransition, NumNLTSInfluencingTrans,
					TransBitLoc,OversampledBitLength);
		x0 = TCentre[i-1][NumTransition-1];
		a = TParam[i-1][NumTransition-1];
	}
	else
	{
		printf("\nERROR: Microtrack() --> Select either LARGE_SPOT or MEDIUM_SPOT \n");
		exit(-1);
	}
	
	/* If temperature is modulated, then reverting back value of x0 to trans centre with the original peak temp.
	   This is important before calculating the readback signal so as to position the current transition in the 
	   same location as a transition without temperature modulation */
	if(TemperatureModulatedFlag == -1)
	{
		x0 = TCentre[i-1][NumTransition-1];
	}

	/* Calculate h(ai,ti) for this microtrack i and add it to the total response p(t) */
	/* Calculate reader sensitivity for the microtrack i */
	rs = exp(-pow((-0.5*(N+1)*delz + i*delz),2)/(2*pow(sigma_r,2)));
	for(j=0 ; j<length ; j++)
	{
		if(j==0)
			x0 = x0 + DeltaX;	/* Simulating jitter. DeltaX = shift in Transition centre.
								If variable SigmaJitter is 0 (no jitter), then DeltaX = 0.
								NOTE that this shift does not change the readback pulse shape.
								*/
		diffx = fabs(x[j] - x0);
		/* 
		v[j] is normalized such that the peak level corresponds to that of the 
		peak level of longitudinal's/perpendicular's isolated readback signal.
		*/
		 v[j] += (*NormalizationFactor) * (1/(float)N) * rs * (Mr[0]*300 + Mr[1]) *(atan((double)(diffx + gr/2)/(a + d)) - atan((double) (diffx - gr/2)/(a + d)));
		
		/* TEST */
		//v[j] += (1/(float)N) * rs * (atan((double)(diffx + gr/2)/(a + d)) - atan((double) (diffx - gr/2)/(a + d)));
	}
	/* Calculating the Magnetization pattern in each track of the medium */
	/*
	M = (double*)malloc(length * sizeof(double));
	for(j = 0 ; j < length ; j++){
		temp = mp->T_Peak*exp(-pow((z0-mp->d),2)/(2*pow(mp->sigma_t,2)))*exp(-pow((x0-mp->c),2)/(2*pow(mp->sigma_t,2))) + 300;
		M[j] = 2 * (mp->Mr[0]*temp + mp->Mr[1]) * atan((x[j]-x0)/a) / PI;
	}
	*/
}
// free(M);

/* Executes if this is a transition that is written basically to calculate PW50, x, T etc */
*PW = 0.0;
if(TransToCalcTransParam == 1)
{
	/* Calculate PW50 */
	min = 1e05;		/* Some large number */
	max = 0;
	/* Determine max v[i]. Assuming only one maximum */
	for(i=0 ; i<length ; i++)
	{
		if(v[i] > max)
		{
			max = v[i];
			index_max = i;
		}
	}
	/* Determine location where v[i] is half the max value to the left of the position where it is maximum */
	for(i=index_max ; i>=0 ; i--)
	{
		if(fabs(v[i]-max/2) < min)
		{
			min = v[i] - max/2;
			left_index_half_max = i;
		}
	}
	/* Determine location where v[i] is half the max value to the rigth of the position where it is maximum */
	min = 1e05;		/* Some large number */
	for(i=index_max ; i<length ; i++)
	{
		if(fabs(v[i]-max/2) < min)
		{
			min = v[i] - max/2;
			right_index_half_max = i;
		}
	}
	/*
	Calculate NormalizationFactor.
	Hamr readback signal is normalized by 'NormalizationFactor' in order to have the same peak 
	level as that of the longitudinal's/perpendicular's isolated readback signal.
	Peak of Long/Perp's readback signal is 2*sqrt(2/pi) if using bipolar bits(+1/-1), or 
	sqrt(2/pi) if using binary bits (0/1).
	*/	
	*NormalizationFactor = (2*sqrt(2/PI))/v[index_max];
	/* Calculate PW50 */
	*PW = x[right_index_half_max] - x[left_index_half_max];
	*PeakAmpPositionIndex = index_max;

	if(DEBUG_MICROTRACK)
	{
		fpDebug = fopen(DebugFileName,"a");
		fprintf(fpDebug,"PW50 = %g\n",*PW);
		fprintf(fpDebug,"Pos of Peak Signal Amp = %g\n",x[index_max]);
		fclose(fpDebug);
	}
}
return;
}


/*------------------------------------- HAMR CHANNEL -------------------------------------*/

#define DEBUG_HAMR 0
void Hamr(struct Mag_Param *mp, struct Physical_Param *pp, struct Reader_Param *rp,int N,
			double *OverSampledInputDataBits, int LengthOfPaddedInputBits,
			double *ReadBackSignal, double OversampledBitLength, int OSR, int NumSectors, double *k, double *rho, double *PW,
			double *x, int length, int *PeakAmpPositionIndex, int NLTSCompensation, int NumNLTSInfluencingTrans,
			int TemperatureVariation, double SigmaTempVariation, double PeakTempTruncValue,
			int WriteHdCrossTrackMov, double MaxWriteHdCrTrMov, double MeanWriteHdCrTrMov, 
			double SigmaJitter, int HMD, double SigmaHMDVariation, int NumARModelCoeff, 
			double *ARModelCoeff, int TemperatureModulation, double ModulatedPeakTemp, 
			double *NormalizationFactor, int TransToCalcTransParam)
 {
	 int i,j,StartIndex,EndIndex,TempIndex;
	 double *v;				/* Readback signal */
	 double x0,z;
	 int NumTransition = 0;	/* Number of Transitions */
	 double **TCentre;		/* Transition Centre location of all microtracks for every transition */
	 double **TParam;		/* Transition Parameter of all microtracks for every transition */
	 int *TransBitLoc;		/* Transition Bit Location */
	 int TemperatureModulatedFlag = 1;
	 double a;
	 double *DeltaD;
	 FILE *fpDebug;
	 char DebugFileName[] = "DebugFile_Hamr.dat";
	 int CountTribits=0;
	 int CountConsecutiveTrans=0;
	 int *LocationOfModulatedBits;
	 int ModBitLocation=0;

	 v = (double*)malloc(length * sizeof(double));
	 if(HMD==1)
		DeltaD = (double*)malloc(sizeof(double));

	 for(i=0 ; i<length ; i++)
	 {
		v[i] = 0;
	 }
	 
	 /* Note that index of Tcentre and Tparam goes from 1 to N */
	 TCentre = (double**)malloc((N+1)*sizeof(double));
	 TParam = (double**)malloc((N+1)*sizeof(double));

	 for(i=0 ; i<=N ; i++)
	 {
		 TCentre[i] = (double*)malloc(sizeof(double));
		 TCentre[i][0] = 0;
		 TParam[i] = (double*)malloc(sizeof(double));
		 TParam[i][0] = 0;
	 }

	 TransBitLoc = (int*)malloc(1*sizeof(int));
	 TransBitLoc[0] = 0;


	 if(DEBUG_HAMR)
	 {
		 LocationOfModulatedBits = (int*)malloc((LengthOfPaddedInputBits/OSR)*sizeof(int));
		 for(i=0;i<LengthOfPaddedInputBits/OSR;i++);
			LocationOfModulatedBits[i]=-1;
		 LocationOfModulatedBits[0]=OverSampledInputDataBits[0];
	 }

	 /* Loop over all input bits */
	 for(i=1;i<LengthOfPaddedInputBits;i++)
	 {/* Loop over all input bits */

		 if(DEBUG_HAMR)
			 if(i%OSR==0)
				LocationOfModulatedBits[i/OSR]=OverSampledInputDataBits[i];

		 /* Indicates a transition */
		 if(OverSampledInputDataBits[i] - OverSampledInputDataBits[i-1] != 0)
		 {
			 	NumTransition ++;			/* These many transitions have occurred so far, including the current one */
				TransBitLoc = (int*)realloc(TransBitLoc,(NumTransition+1)*sizeof(int));
				TransBitLoc[NumTransition] = i; /* NOTE: Index of TransLoc is from 1 and NOT 0 */

				/* HMD Variation */
				if(HMD == 1)
					DeltaD = (double*)realloc(DeltaD,NumTransition*sizeof(double));

/*				if(TemperatureModulation == 1 && NumSectors >=0 && TransToCalcTransParam==0)
				{
					mp->T_Peak = ModulatedPeakTemp;
				}
*/				
				/* Temperature Modulation: detecting 3 or more consecutive transition */
				if(TemperatureModulation == 1 && NumSectors >= 0 && TransToCalcTransParam==0)
				{
					/* Assuming that since the data is pre and post padded with zeroes,
					 * there wont be any tribits at the beginning or end 
					 */
					if(i >= 2*OSR && i<LengthOfPaddedInputBits-3*OSR)
					{
						/* This transition is preceded and succeeded by atleast one transition */
						if(OverSampledInputDataBits[i-OSR]-OverSampledInputDataBits[i-OSR-1] != 0 &&
							OverSampledInputDataBits[i+OSR]-OverSampledInputDataBits[i+OSR-1] != 0)
						{
							CountConsecutiveTrans++;	/* Counts number of consecutive transitions
														 * if 3 or more consecutive transitions occur */
							if(NumSectors >= 0 && (CountConsecutiveTrans%3 == 2))	/* Modulate only the middle bit of every consecutive tribits */
							{
								ModBitLocation = i;
								CountTribits++;
//								if(CountTribits==50)
//									mp->T_Peak = ModulatedPeakTemp;
//								TemperatureModulatedFlag = -1;
								if(DEBUG_HAMR)
									LocationOfModulatedBits[i/OSR] = -LocationOfModulatedBits[i/OSR];
							}
							else
							{
								if((i == ModBitLocation+OSR) || (OverSampledInputDataBits[i+OSR]-OverSampledInputDataBits[i+OSR-1] != 0 &&
									OverSampledInputDataBits[i+2*OSR]-OverSampledInputDataBits[i+2*OSR-1] != 0))
								{
//									NumTransition--;
									if(DEBUG_HAMR)
										LocationOfModulatedBits[i/OSR] = LocationOfModulatedBits[i/OSR-1];
//									continue;
								}
							}
						}/* Atleast two transitions precede this transition */
						else if(OverSampledInputDataBits[i-OSR]-OverSampledInputDataBits[i-OSR-1] != 0 &&
							OverSampledInputDataBits[i-2*OSR]-OverSampledInputDataBits[i-2*OSR-1] != 0)
						{
							CountConsecutiveTrans++;
							if(CountConsecutiveTrans%3==0)
							{
//								NumTransition--;
								CountConsecutiveTrans = 0;
								if(DEBUG_HAMR)
									LocationOfModulatedBits[i/OSR] = LocationOfModulatedBits[i/OSR-1];
//								continue;
							}
							CountConsecutiveTrans = 0;

						}/* This transition is followed by atleast two transitions */
						else if(OverSampledInputDataBits[i+OSR]-OverSampledInputDataBits[i+OSR-1] != 0 &&
							OverSampledInputDataBits[i+2*OSR]-OverSampledInputDataBits[i+2*OSR-1] != 0)
						{
							CountConsecutiveTrans++;
							if(CountConsecutiveTrans%3==1)
							{
//								NumTransition--;
								mp->T_Peak = ModulatedPeakTemp;
								/* TEMP CODE */
								TemperatureModulatedFlag = -1;
								if(DEBUG_HAMR)
									LocationOfModulatedBits[i/OSR] = LocationOfModulatedBits[i/OSR-1];
//								continue;
							}
						}
						else
						{
							CountConsecutiveTrans = 0;
						}
					}
				}
				Microtrack(mp,pp,rp,N,x,v,PW,length,PeakAmpPositionIndex,NumTransition,TCentre,TParam,k,rho,TransBitLoc,
						  NumNLTSInfluencingTrans, OversampledBitLength, OSR, NumSectors, TemperatureVariation, SigmaTempVariation, 
						  PeakTempTruncValue,WriteHdCrossTrackMov, MaxWriteHdCrTrMov, MeanWriteHdCrTrMov,
						  SigmaJitter, HMD, SigmaHMDVariation, NumARModelCoeff, ARModelCoeff, DeltaD, NormalizationFactor,TransToCalcTransParam,
						  TemperatureModulatedFlag);
				
				/* Reverting back to the original peak temperature, if Temperature Modulation is ON */
				/* TEMP CODE */
				TemperatureModulatedFlag = 1;

				if(TemperatureModulation == 1)
				{
					mp->T_Peak = mp->Orig_T_Peak;
				}
		 }

		 /* Readback signal is +ve */
		 if(OverSampledInputDataBits[i] - OverSampledInputDataBits[i-1] == +2*TemperatureModulatedFlag)
		 {
			 StartIndex = (*PeakAmpPositionIndex) > i ? (*PeakAmpPositionIndex) - i : 0;
			 EndIndex = (length - (*PeakAmpPositionIndex)) > (LengthOfPaddedInputBits - i) ? ((*PeakAmpPositionIndex) + LengthOfPaddedInputBits - i) : length;
			 TempIndex = i - ((*PeakAmpPositionIndex) - StartIndex);
			 for(j=StartIndex ; j<EndIndex ; j++)
			 {
				 ReadBackSignal[TempIndex] += v[j];
				 TempIndex++;
			 }
			 for(j = 0 ; j < length ; j++)
			 {
				v[j] = 0;
			 }
		 /* Readback signal is -ve */
		 }
		 else if(OverSampledInputDataBits[i] - OverSampledInputDataBits[i-1] == -2*TemperatureModulatedFlag)
		 {
			 StartIndex = (*PeakAmpPositionIndex) > i ? (*PeakAmpPositionIndex) - i : 0;
			 EndIndex = (length - (*PeakAmpPositionIndex)) > (LengthOfPaddedInputBits - i) ? ((*PeakAmpPositionIndex) + LengthOfPaddedInputBits - i) : length;
			 TempIndex = i - ((*PeakAmpPositionIndex) - StartIndex);
			 for(j=StartIndex ; j<EndIndex ; j++)
			 {
				 ReadBackSignal[TempIndex] -= v[j];
				 TempIndex++;
			 }
			 for(j = 0 ; j < length ; j++)
			 {
				v[j] = 0;
			 }
		 }
		 TemperatureModulatedFlag = 1;
	 }/* Loop over all input bits */

	 if(DEBUG_HAMR==1 && NumSectors >= 0 && TransToCalcTransParam==0)// && NumNLTSInfluencingTrans > 0)
	 {
		 fpDebug = fopen(DebugFileName,"a");

		 for(i=0;i<LengthOfPaddedInputBits/OSR;i++)
			 fprintf(fpDebug," %d",LocationOfModulatedBits[i]);
		 fprintf(fpDebug,"\n");

/*		 for(i=0;i<NumTransition;i++)
		 {
			 for(j=0;j<N;j++)
			 {
				z = pp->wt/2 - (pp->wt/N*(j) + pp->wt/N/2);
				fprintf(fpDebug,"%d %d %g %g %g\n",i+1,j+1,z,TParam[j][i],TCentre[j][i]);
			 }
			 fprintf(fpDebug,"\n");
		 }
*/
		 //fprintf(fpDebug,"\nPosition of Peak signal = %g\n",x[*PeakAmpPositionIndex]);
		 //fprintf(fpDebug,"PW50 = %g\n\n",*PW);

		 fclose(fpDebug);
	 }

	 free(v);
	 for(i=0;i<=N;i++)
	 {
		 free(TCentre[i]);
		 free(TParam[i]);
	 }
	 free(TCentre);
	 free(TParam);
	 free(TransBitLoc);
	 if(HMD==1)
		 free(DeltaD);
	 return;
 }


/*------------------------- LONGITUDINAL/PERPENDICULAR CHANNEL --------------------------*/

void LongPerp(double *OSPadEncodedBits, int OSSectorLength, double *OSChannelOutput, double *ReadChannelCoeff,
			int NumChannelTaps, double *TimeIndex, double ND, double SigmaJitter, double SigmaPulseBroad,
			int OSR, char *ChannelType, int j)
{
	/* 
	Converting OSPadEncodedBits to a double array, because NonCausalFIR() 
	expects a double array (and not int)
	*/

	/* Channel with no media noise (AWGN) only */
	if(SigmaJitter == 0 && SigmaPulseBroad == 0)
	{
		NonCausalFIR(OSPadEncodedBits, OSSectorLength, OSChannelOutput,ReadChannelCoeff,NumChannelTaps);
	}
	/* Channel with media noise (jitter + pulse broadening effect) */
	else
	{
		/* Non-Adaptive mode */
		if(j>=0)
		{
			MediaNoiseFilter(OSPadEncodedBits, OSSectorLength, OSChannelOutput, ReadChannelCoeff,
								NumChannelTaps, TimeIndex, ND, SigmaJitter, SigmaPulseBroad,OSR,
								ChannelType);
		}
		/* Adaptive Mode */
		else
		{
			/* Making SigmaJitter and SigmaPulseBroad zero */
			MediaNoiseFilter(OSPadEncodedBits, OSSectorLength, OSChannelOutput, ReadChannelCoeff,
								NumChannelTaps, TimeIndex, ND, 0.0, 0.0,OSR, ChannelType);
		}
	}
}

/*-------------------------------------- MAGNETIC DISK ---------------------------------------*/

#define DEBUG 1
#define ERROR_EVENTS 0
#define UNCODED 1

/* This is the MAIN simulator function */
void main(int argc, char *argv[])
{

/****************************** USER INPUT PARAMETERS *******************************/

/*** General Parameters to be set for any kind of simulation ***/
double SNR[] = {21};//{28.0,29.0,30.0,31.0,32.0,33.0,34.0,35.0,36.0,37.0};
int MaxNumSectors = 10;//3000;
int MinNumSectors = 10;
int MaxNumBitErr = 100;
int OSR = 10;
int SectorLength = 4096;
int PrePaddingLength = 20;
int PostPaddingLength = 20;

/*** Channel Options ***/
char ChannelType[] = "Longitudinal";		/* OPTIONS : Longitudinal OR Perpendicular OR Hamr */
double UserDensity = 2.5;				/* Density before coding */
/** Longitudinal and Perpendicular channel specific parameters **/
int NumChannelTaps = 201;			/* Number of Channel filter taps */
/** HAMR channel specific parameters are specified at the end of all variable declarations **/

/*** Noise Parameters ***/
/** Media noise specific parameters **/
double SigmaJitter = 0.0;			/* Standard deviation of jitter noise as a % of bit period (T) */
double SigmaPulseBroad = 0.0;		/* Standard deviation of pulse broadening noise as a % of bit period (T) */

/*** LPF specific parameters ***/
int NumLPFTaps = 201;				/* Number of Low Pass Filter taps */

/*** Equalizer Options ***/
char EqualizerType[] = "GPRTarget";	/* OPTIONS: FixedPRTarget OR GPRTarget */
int NumEqTaps = 21;						/* Number of PR/GPR Filter taps */
int NumEqSectors = 10;					/* Number of sectors used for adapting equalizer */

/* PR Fixed Target specific parameters */
double PRImpRes[] = {1,1,-1,-1};//{1,2,0,-2,-1};										
										/* PR Impulse Response values 
										Note: include commas between values. DO NOT include zeroes
										at the beginning or the end.
										*/
/* GPR specific parameters */
int GPRTargetLength = 5;				/* Length of the Generalized Partial Response Target */

/*** Detector Options ***/
char DetectorType[] = "Viterbi";		/* OPTIONS: Viterbi OR SOVA */
int ViterbiDelay = 20;					/* Delay in detection for viterbi AND sova */

/*** Output File Names ***/
char SimulationSummary[] = "./SimulationSummary.dat";
char ErrorEvents[] = "./ErrorSectors.dat";

/*************************************************************************/
/*************************************************************************/
/*************************************************************************/

/*** Simulator internal variables. DO NOT change these ***/
long int idum1 = -500, idum2 = -600;
int i, j, k, NumSectors=0, N, InvalidCW;
int CountNRZDecErr,CountNRZIDecErr,CountNRZVitErr,CountNRZIVitErr;
int NumDecSecErr,NumVitSecErr;
int NumNRZDecErr,NumNRZIDecErr,NumNRZVitErr,NumNRZIVitErr;
clock_t start = clock();
int MSECheckFlag=1, StartFlag=1;
int PRImpResLength, PaddedSectorLength, OSSectorLength, SNRArrayLength, UserSectorLength;
int *UserBits, *EncodedBits, *PaddedBits, *OSBits;
int *ViterbiHardOutput, *DecodedOutput;
int *States,*ModulationValues, **ValidIncomingTrans, NumberOfStates;
int NumCWSector, CodewordLength, UserwordLength;
char DebugFile[] = "DebugFile.dat";
double *GPRTarget, *GPRConcatChannelOutput, *GPRConcatChannelInput;
double *EqCoeff;
double *PaddedBipolarBits, *OSBipolarBits;
double *TimeIndex, *ViterbiSoftOutput, *ReadChannelCoeff, *ChannelOutput, *LPFOutput, *DSOutput, *EqualizedOutput;
double *ber, *ser;
double CurrentSNR, NoiseSigma, CutOff, CodeRate, T, ND, Vp;
FILE *fpSimulationSummary=NULL, *fpDebugFile=NULL, *fpErrorEvents=NULL, *fpARModelCoeff=NULL;
struct Mag_Param *mp;	
struct Physical_Param *pp;
struct Reader_Param *rp;
int TemperatureVariation, WriteHdCrossTrackMov, HMD, NumARModelCoeff;
int TemperatureModulation;
double ModulatedPeakTemp; 
double SigmaTempVariation, PeakTempTruncValue, MaxWriteHdCrTrMov, MeanWriteHdCrTrMov;
double SigmaHMDVariation = 0;
double *ARModelCoeff;
char ARModelCoeffFileName[50]="";
double EnergyWithPW1,EnergyWithPW2,PW1,PW2,Vp1,Vp2;
int NSymbols,NNeighbors,m,n,*ModulatedTribitIndex;
double *SampleValues;


int NumNLTSInfluencingTrans = 0;	/* Number of past transitions that contributes to NLTS */
int *PeakAmpPositionIndex;			/* Index of x where readback signal is at its peak */
double NormalizationFactor = 1;		/* Maximum readback signal level */
int XLength;						/* Length of x */
double PW = 0.0;					/* PW50 */
double *X;							/* Downtrack location where readback voltage will be calculated */
double *K, *Rho;					/* NLTS parameters K, Rho */
double *KDummy, *RhoDummy;			/* NLTS parameters K, Rho set to 0 (used when media noise should be excluded)*/
double OverSampledBitLength = 0.0;	/* Bit length in the oversampled domain */
double NLTSCompensation = 0.0;		/* Compensation in 'nm' for NLTS */
double temp;

mp = (struct Mag_Param *) malloc(sizeof(struct Mag_Param));
pp = (struct Physical_Param *) malloc(sizeof(struct Physical_Param));
rp = (struct Reader_Param *) malloc(sizeof(struct Reader_Param));
/*** END of Simulator internal variables ***/

/****** HAMR Specific Parameters *******/
N = 16;						/* Number of microtracks */

/* Magnetization/Thermal Profile Parameters */
mp->sigma_t = 90;			/* Sigma of T-Profile in nm */
mp->T_Peak = 350;			/* Peak Temperature in centigrade */
mp->c = 0.0;				/* Downtrack location of peak temperature w.r.t gap centreline in nm */
mp->d = 0.0;				/* Cross track location of peak temperature w.r.t track centre in nm */
mp->z0 = 0.0;				/* T_calc = T_peak * exp(-z0/(2*sigma_t^2)) */
mp->Hc[0] = -1000;//-2000;			/* Dependence of coercivity on Temperature in A/m. Hc = aT + b => {a,b} */
mp->Hc[1] = 1600000;
mp->Mr[0] = -600;//-1200;			/* Dependence of remnant magnetization on Temperature in A/m. Mr = aT + b => {a,b} */
mp->Mr[1] = 1200000;
mp->S[0] = 0.0003;				/* Dependence of coercive squareness on Temperature. S = aT + b => {a,b} */
mp->S[1] = 0.5;
mp->Hg =  16e05;			/* Deep gap field in A/m */

/* Physical Parameters */
pp->g = 100;				/* Write/Head gap in nm */
pp->d = 19;					/* Medium spacing(in nm) or the space between the bottom of pole and the surface of the medium */
pp->t = 2;					/* Medium thickness in nm */
pp->y = pp->d + pp->t/2;	/* Distance from bottom of pole to the centre of the medium in nm */
pp->wt = 160;				/* Width of the track */

/* GMR Reader Parameters */
rp->C = 1;						/* Constant which describes the physical attributes of GMR reader */
rp->gr = 5;						/* Spacing between the shield and the GMR element (nm) */
rp->tr = 1;						/* Width of the GMR element (nm) */
rp->wr = 1000;					/* Width of the reader (in Z direction, i.e. crosstrack direction) */
rp->sigma_r = 1000;				/* Sigma of Reader sensitivity function, which is Gaussian (nm) */

/* NLTS Parameters */
NumNLTSInfluencingTrans = 0;
K = (double*)malloc(N*sizeof(double));		/* NLTS = K/B^Rho, where B is the distance between two trans */
Rho = (double*)malloc(N*sizeof(double));
for(i=0;i<N;i++)							/* K, Rho for each track */
{
	K[i] = 4.0e03;
	Rho[i] = 2.0;
}
/* Temperature Modulation Parameters */
TemperatureModulation = 0;	/* Switch ON(1) or OFF(0) the peak temperature modulation */
ModulatedPeakTemp = 450;	/* Peak Temperature to be used for 3 or more consecutive transitions */

 
/* Peak Temperature Variation Parameters */
TemperatureVariation = 0;	/* Switch ON(1) or OFF(0) the variation in peak temperature effect */
SigmaTempVariation = 2;		/* Peak Temp is modeled as Truncated Gaussian with sigma = SigmaTempVariation % of Peak Temp */
PeakTempTruncValue = 20;	/* Peak temperature is varied from < PeakTemp-PeakTempTruncValue to PeakTemp+PeakTempTruncValue > 
							   according to a Gaussian profile */

/* Write Head Cross-Track Movement Parameters */
WriteHdCrossTrackMov = 0;	/* Switch ON(1) or OFF(0) the write head cross track movement */
MaxWriteHdCrTrMov = 20;		/* Maximum write head cross track movement as a % of track width */
MeanWriteHdCrTrMov = 10;	/* Mean cross track movement of write head as % of track width */


/* Variation in Head-Medium Distance */
HMD = 0;						/* Switch ON(1) or OFF(0) the variation in Head-Medium Distance effect */
SigmaHMDVariation = sqrt(2.0);	/* The HMD Variation is modeled as an AR process. This variable represents the 
								 * sigma of the Gaussian white process of the AR model */
sprintf(ARModelCoeffFileName,"%s","ARModelCoeff.dat"); /* File that contains the coeff of the AR model */

if(HMD==1)
{
	/** Reading AR Coefficients from the above specified file **/
	fpARModelCoeff = fopen(ARModelCoeffFileName,"r");
	if(ARModelCoeffFileName==NULL)
	{
		printf("\nERROR: Main -> Cannot open file %s\n",ARModelCoeffFileName); 
		exit(-1);
	}
	fscanf(fpARModelCoeff,"%d\n",&NumARModelCoeff);
	ARModelCoeff = (double*)malloc(NumARModelCoeff*sizeof(double));
	for(i=0;i<NumARModelCoeff;i++)
	{
		fscanf(fpARModelCoeff,"%lg\n",&(ARModelCoeff[i]));
	}
	fclose(fpARModelCoeff);
}

/*********************************************************************************************/
/*********************************************************************************************/
/*---------------------- USER NEED NOT CHANGE ANYTHING BEYOND THIS LINE ---------------------*/
/*********************************************************************************************/
/*********************************************************************************************/

/* Open all necessary files  for reading and writing */
fpSimulationSummary = fopen(SimulationSummary, "a");
if(fpSimulationSummary == NULL)
{
	printf("\nERROR: Main -> Cannot create file %s\n",SimulationSummary);
	exit(-1);
}

if(ERROR_EVENTS)
{
	fpErrorEvents = fopen(ErrorEvents,"a");
	if(fpErrorEvents == NULL)
	{
		printf("\nERROR: Main -> Cannot create file %s\n",ErrorEvents);
		exit(-1);
	}
}

/** Calculating SectorLength **/
if(UNCODED)
{
	UserwordLength = 1;
	CodewordLength = 1;
}else
{
	/* Hard coding */
	UserwordLength = 4;
	CodewordLength = 5;
}
CodeRate = (double)UserwordLength/(double)CodewordLength; 
/* 
Calculate actual sector length. 
Sector length = max(Codeword Length, User specified sector length)
*/
if(UNCODED)
{
	if(SectorLength <= CodewordLength)
	{
		SectorLength = CodewordLength;
		NumCWSector = 1;
		UserSectorLength = UserwordLength;
	}else
	{
		NumCWSector = SectorLength/CodewordLength;
		/* A sector should contain integer number of codewords */
		SectorLength = NumCWSector*CodewordLength;
		UserSectorLength = NumCWSector*UserwordLength;
	}
}else
{
	/* Hard coded for modulation code 6/7 MTR(2;8) code */
	//UserSectorLength = (SectorLength/6)*6 + 1;
	//SectorLength = (UserSectorLength-1)/CodeRate + 1;	/* Sector length after coding */

	/* Hard coded for modulation code 8/9 TMTR(2/3;11) code */
	UserSectorLength = (SectorLength/4)*4 + 1;
	SectorLength = (UserSectorLength-1)/CodeRate + 1;	/* Sector length after coding */
}

/* Adjusting Symbol Density due to coding */
//printf("\n ***** Running TEST Code ***** \n");
ND = UserDensity/CodeRate;

/*---- Initialization ----*/
UserBits = (int*)malloc(UserSectorLength*sizeof(int));
for(i=0;i<UserSectorLength;i++){
	UserBits[i] = 0;
}
EncodedBits = (int*)malloc(SectorLength*sizeof(int));
for(i=0;i<SectorLength;i++){
	EncodedBits[i] = 0;
}
PaddedSectorLength = SectorLength + PrePaddingLength + PostPaddingLength;
PaddedBits = (int*)malloc(PaddedSectorLength*sizeof(int));
PaddedBipolarBits = (double*)malloc(PaddedSectorLength*sizeof(double));
for(i=0;i<PaddedSectorLength;i++){
	PaddedBits[i] = 0;
	PaddedBipolarBits[i] = 0;
}
OSSectorLength = PaddedSectorLength*OSR;
OSBits = (int*)malloc(OSSectorLength*sizeof(int));
for(i=0;i<OSSectorLength;i++){
	OSBits[i] = 0;
}
OSBipolarBits = (double*)malloc(OSSectorLength*sizeof(double));
for(i=0;i<OSSectorLength;i++){
	OSBipolarBits[i] = 0;
}
ChannelOutput = (double*)malloc((OSSectorLength+ NumChannelTaps/2)*sizeof(double));
for(i=0;i<(OSSectorLength+ floor(NumChannelTaps/2));i++){
	ChannelOutput[i] = 0.0;
}
LPFOutput = (double*)malloc((OSSectorLength + NumLPFTaps - 1)*sizeof(double));
for(i=0;i<(OSSectorLength+NumLPFTaps-1);i++){
	LPFOutput[i] = 0.0;
}
DSOutput = (double*)malloc(PaddedSectorLength*sizeof(double));
for(i=0;i<PaddedSectorLength;i++){
	DSOutput[i] = 0.0;
}
EqualizedOutput = (double*)malloc((PaddedSectorLength + NumEqTaps/2)*sizeof(double));
for(i=0;i<(PaddedSectorLength+NumEqTaps/2);i++){
	EqualizedOutput[i] = 0.0;
}
ViterbiHardOutput = (int*)malloc(PaddedSectorLength*sizeof(int));
for(i=0;i<PaddedSectorLength;i++){
	ViterbiHardOutput[i] = 0;
}
ViterbiSoftOutput = (double*)malloc(PaddedSectorLength*sizeof(double));
for(i=0;i<PaddedSectorLength;i++){
	ViterbiSoftOutput[i] = 0.0;
}
DecodedOutput = (int*)malloc(UserSectorLength*sizeof(int));
for(i=0;i<UserSectorLength;i++){
	DecodedOutput[i] = 0;
}
ReadChannelCoeff = (double*)malloc(NumChannelTaps*sizeof(double));
TimeIndex = (double*)malloc(NumChannelTaps*sizeof(double));
for(i=0;i<NumChannelTaps;i++)
{
	ReadChannelCoeff[i] = 0.0;
	TimeIndex[i] = 0.0;
}
EqCoeff = (double*)malloc(NumEqTaps*sizeof(double));
for(i=0;i<NumEqTaps;i++)
{
	EqCoeff[i] = 0.0;
}
GPRTarget = (double*)malloc(GPRTargetLength*sizeof(double));
for(i=0;i<GPRTargetLength;i++)
{
	GPRTarget[i]=0.0;
}
GPRConcatChannelInput = (double*)malloc(NumEqSectors*PaddedSectorLength*sizeof(double));
GPRConcatChannelOutput = (double*)malloc(NumEqSectors*PaddedSectorLength*sizeof(double));

if(!strcmp(EqualizerType,"FixedPRTarget"))
{
	PRImpResLength = sizeof(PRImpRes)/sizeof(double);
}else if(!strcmp(EqualizerType,"GPRTarget"))
{
	PRImpResLength = GPRTargetLength;
}
SNRArrayLength = sizeof(SNR)/sizeof(double);
ber = (double*)malloc(SNRArrayLength*sizeof(double)); /* is it same as sizeof(SNR) ? - Not sure! */
ser = (double*)malloc(SNRArrayLength*sizeof(double)); /* is it same as sizeof(SNR) ? - Not sure! */

/* Temporary Code */
//printf("WARNING: main()-->Running temporary code\n");
States = (int*)malloc(1024*sizeof(int));
ModulationValues = (int*)malloc(1024*sizeof(int));
ValidIncomingTrans = (int**)malloc(1024*sizeof(int));
for(i=0;i<1024;i++)
{
	ValidIncomingTrans[i]=(int*)malloc(3*sizeof(int));
	ValidIncomingTrans[i][0]=-1;ValidIncomingTrans[i][1]=-1;ValidIncomingTrans[i][2]=-1;
	States[i]=0;
	ModulationValues[i]=0;
}
NumberOfStates = 0;

/*---- Determining channel coefficients ----*/
/* 
Determining the time index, according to NumChannelTaps and OSR, for which the channel
taps will be calculated.
*/
TimeIndex[0] = floor(NumChannelTaps/2)/(double)OSR;
for(i=1;i<NumChannelTaps;i++)
{
	TimeIndex[i] = TimeIndex[i-1] - 1/(double)OSR;	/* Be wary of problems due to precision that
													might occur here. For ex: 0 might not be a
													TimeIndex.
													*/
}
if(!strcmp(ChannelType,"perpendicular") || !strcmp(ChannelType,"Perpendicular"))
{
	Vp = 0.5*pow(PI/(2*log(2)),0.25);
	for(i=0;i<NumChannelTaps;i++)
	{
		ReadChannelCoeff[i] = Vp * 2/ND * sqrt(log(2)/PI) * exp(-log(2)*4*pow(TimeIndex[i],2)/pow(ND,2));
		ReadChannelCoeff[i] = ReadChannelCoeff[i]/(double)OSR;
	}
}else if(!strcmp(ChannelType,"longitudinal") || !strcmp(ChannelType,"Longitudinal"))
{
	Vp = sqrt(2.0/PI);
	for(i=0;i<NumChannelTaps;i++)
	{
		ReadChannelCoeff[i] = Vp * -8*TimeIndex[i]/pow(ND,2) * 1/pow((1+4*pow(TimeIndex[i],2)/pow(ND,2)),2);
		ReadChannelCoeff[i] = ReadChannelCoeff[i]/(double)OSR;
	}
}else if(!strcmp(ChannelType,"Hamr") || !strcmp(ChannelType,"HAMR") || !strcmp(ChannelType,"hamr"))
{
	/* Nothing needs to be done, since channel output is calculated numerically */
}else
{
	printf("ERROR: -> Main() -> Incorrect channel type %s. Program terminated\n",ChannelType);
	exit(-1);
}

/** 
Unlike, Perpendicular and longitudinal channel, Hamr channel is numerical. Therefore,
channel characteristics like PW50, T etc are not known initially. In order to determine these,
a single transition is written and its readback signal is generated.
**/
if(!strcmp(ChannelType,"Hamr") || !strcmp(ChannelType,"hamr") || !strcmp(ChannelType,"HAMR"))
{/* Initial Hamr channel parameters */
	/* Making one transition */
	for (i= 0; i<= OSSectorLength/2; i++)
	{
		OSBipolarBits[i]= -1;
	}
	for (i= OSSectorLength+1; i < OSSectorLength; i++)
	{
		OSBipolarBits[i]= +1;
	}
	PeakAmpPositionIndex = (int*)malloc(sizeof(int));
	PeakAmpPositionIndex[0] = 0;
	/* Excluding media noise for this transition */
	KDummy = (double*)malloc(N*sizeof(double));
	RhoDummy = (double*)malloc(N*sizeof(double));
	for(i=0;i<N;i++)
	{
		KDummy[i] = 0.0;
		RhoDummy[i] = 0.0;
	}
	mp->Orig_T_Peak = mp->T_Peak;
	mp->Orig_d = mp->d;
	/* Generating downtrack locations where readback levels will be calculated. 
	* These are sampled at 1nm in order to get an accurate PW50 value and also
	* to capture the peak accurately.
	*/
	/* Depending on the PW50, these locations will be optimized for other transitions */
	/* Now, x ranges from -1000 to 1000 nm sampled at 1 nm */

	X = (double*)malloc(sizeof(double));
	for(i=0;i<=2000;i++){
		X = (double*)realloc(X,(i+1)*sizeof(double));
		X[i] = -1000 + i;
	}
	XLength = 2001;
	Hamr(mp, pp, rp, N, OSBipolarBits, OSSectorLength, ChannelOutput, 
		OverSampledBitLength, OSR, NumSectors, KDummy, RhoDummy, &PW, X, XLength, PeakAmpPositionIndex, NLTSCompensation, 
		NumNLTSInfluencingTrans,TemperatureVariation, SigmaTempVariation, PeakTempTruncValue,
		WriteHdCrossTrackMov, MaxWriteHdCrTrMov, MeanWriteHdCrTrMov, SigmaJitter, HMD, SigmaHMDVariation, 
		NumARModelCoeff, ARModelCoeff, TemperatureModulation, ModulatedPeakTemp, &NormalizationFactor,1);

	OverSampledBitLength = PW/(ND*OSR);
	temp = X[(*PeakAmpPositionIndex)];
	free(X);

	/* The new 'x' will be 10*PW50 either side of the peak value sampled according to OverSampledBitLength */
	XLength = 2*((int)(10*PW/OverSampledBitLength)+1)+1;
	X = (double*)malloc(XLength*sizeof(double));
	X[((XLength-1)/2)] = temp;
	for(i=((XLength-1)/2)+1;i<XLength;i++)
	{
		X[i] = X[i-1] +  OverSampledBitLength;
	}
	for(i=((XLength-1)/2)-1;i>=0;i--)
	{
		X[i] = X[i+1] -  OverSampledBitLength;
	}
	PeakAmpPositionIndex[0] = ((XLength-1)/2);
	
/*	X = (double*)malloc(sizeof(double));
	XLength = 2*((int)(8*PW/OverSampledBitLength)+1)+1;
	X[0] = temp - OverSampledBitLength * ((XLength-1)/2);
	for(i=1;i<XLength;i++)
	{
		X = (double*)realloc(X,(i+1)*sizeof(double));
		X[i] = X[i-1] + OverSampledBitLength;
	}
	PeakAmpPositionIndex[0] = ((XLength-1)/2);
*/
	/* TEST */
	printf("PW50 = %g\n",PW);
	printf("Normalization Factor = %g\n",NormalizationFactor);


	/* Set the K and Rho values here if transition noise needs to be added to Hamr Channel */
/* for(i=0;i<N;i++)
	{
	K[i] = 1.5e06;
	Rho[i] = 2.7;
	}
*/

}/* Initial Hamr channel parameters */

/* Write all system parameters to the output file */
fprintf(fpSimulationSummary,"Simulation Started @ %s on %s\n", __TIME__, __DATE__);
fprintf(fpSimulationSummary,"File Executed : %s\n\n", __FILE__);

fprintf(fpSimulationSummary,"Channel Type = %s\n",ChannelType);
fprintf(fpSimulationSummary,"Equalizer Type = %s\n",EqualizerType);
fprintf(fpSimulationSummary,"PR Impulse Response: [ ");
for(i=0;i<PRImpResLength;i++)
	fprintf(fpSimulationSummary,"%g ",PRImpRes[i]);
fprintf(fpSimulationSummary,"]\n");
fprintf(fpSimulationSummary,"User Density = %g\n",UserDensity);
fprintf(fpSimulationSummary,"ND = %g\n",ND);

if(!strcmp(ChannelType,"Hamr") || !strcmp(ChannelType,"hamr") || !strcmp(ChannelType,"HAMR"))
{
	fprintf(fpSimulationSummary,"\n-------THERMAL PROFILE-------\n");
	fprintf(fpSimulationSummary,"Sigma of T = %g nm\n",mp->sigma_t);
	fprintf(fpSimulationSummary,"Peak temperature = %g K\n",mp->T_Peak);
	fprintf(fpSimulationSummary,"Shift in peak temperature = %g nm\n",mp->c);

	fprintf(fpSimulationSummary,"\n-------MAGNETIC PROPERTIES-------\n");
	fprintf(fpSimulationSummary,"Coercivity = %gT + %g A/m\n",mp->Hc[0],mp->Hc[1]);
	fprintf(fpSimulationSummary,"Remnant magnetization = %gT + %g A/m\n",mp->Mr[0],mp->Mr[1]);
	fprintf(fpSimulationSummary,"Coercive squareness = %gT + %g\n",mp->S[0],mp->S[1]);
	fprintf(fpSimulationSummary,"Deep gap field = %g A/m\n",mp->Hg);

	fprintf(fpSimulationSummary,"\n-------PHYSICAL PARAMETERS-------\n");
	fprintf(fpSimulationSummary,"Write head gap = %g nm\n",pp->g);
	fprintf(fpSimulationSummary,"Distance from bottom of pole to medium centre = %g nm\n",pp->y);
	fprintf(fpSimulationSummary,"Width of the track = %g nm\n",pp->wt);
	fprintf(fpSimulationSummary,"Number of microtracks = %d\n",N);

	fprintf(fpSimulationSummary,"\n-------READER PARAMETERS-------\n");
	fprintf(fpSimulationSummary,"Read head gap = %g nm\n",rp->gr);
	fprintf(fpSimulationSummary,"Reader width = %g nm\n",rp->wr);
	fprintf(fpSimulationSummary,"Sigma of Reader sensitivity = %g nm\n",rp->sigma_r);
	fprintf(fpSimulationSummary,"XStart = %g nm\n",X[0]);
	fprintf(fpSimulationSummary,"XEnd = %g nm\n",X[XLength-1]);
	fprintf(fpSimulationSummary,"Resolution of ReadBack signal = %g nm\n",OverSampledBitLength);

	fprintf(fpSimulationSummary,"\n-------MISC-------\n");
	fprintf(fpSimulationSummary,"PW50 = %g nm\n",PW);
	fprintf(fpSimulationSummary,"Number of NLTS Influencing Transitions = %d\n",NumNLTSInfluencingTrans);
	fprintf(fpSimulationSummary,"K = %g\t Rho = %g\n\n",K[0],Rho[0]);
	fprintf(fpSimulationSummary,"Temperature Variation Effect = %d (1->ON,0->OFF)\n",TemperatureVariation);
	fprintf(fpSimulationSummary,"Sigma of Peak Temperature Variation = %g%% of peak temperature\n",SigmaTempVariation);
	fprintf(fpSimulationSummary,"Peak Temperature Varies between < %gC to %gC >\n",mp->Orig_T_Peak-PeakTempTruncValue,mp->Orig_T_Peak+PeakTempTruncValue);
	fprintf(fpSimulationSummary,"Cross Track Movement of Write Head = %d (1->ON,0->OFF)\n",WriteHdCrossTrackMov);
	fprintf(fpSimulationSummary,"Mean cross track movement of write head = %g%% of track width\n",MeanWriteHdCrTrMov);
	fprintf(fpSimulationSummary,"Maximum write head cross track movement = %g%% of track width\n",MaxWriteHdCrTrMov);
}
fclose(fpSimulationSummary);

for(j = -1 ; j < SNRArrayLength ; j++)
{/* Loops over all SNR values */

	/* j = -1 implies equalizer is in adaptive mode */
	if(j == -1)
	{
		/* Set current snr to some big value so as to avoid including any electronic noise while equalizing */
		CurrentSNR = 50;
	}else
	{
		CurrentSNR = SNR[j];
		ber[j] = 0; 
		ser[j] = 0;
		if(ERROR_EVENTS)
		{
			fprintf(fpErrorEvents,"%g\n",CurrentSNR);
		}
	}
	NumSectors = 0;
	CountNRZDecErr=0; CountNRZIDecErr=0; CountNRZVitErr=0; CountNRZIVitErr=0;
	NumDecSecErr=0; NumVitSecErr=0;
	NumNRZDecErr=0; NumNRZIDecErr=0; NumNRZVitErr=0; NumNRZIVitErr=0;

	if(j==-1)
	{
		/* -ve NumSectors indicate equalizer is in adaptive mode */
		NumSectors = -NumEqSectors;
		if(NumSectors == 0)
			continue;
	}

	/************* MAIN LOOP **************/
	while(NumSectors < MaxNumSectors && (NumSectors <= MinNumSectors || NumNRZIVitErr <= MaxNumBitErr))
	{/* sector count loop */

		/* If Equalizer has been adapted, change the SNR to user specified one */
		if(j==-1 && NumSectors == 0)
		{
			j++; /* NOTE: Changing FOR loop index. Hopefully, should not give any problems! */
			CurrentSNR = SNR[j];
			if(ERROR_EVENTS)
			{
				fprintf(fpErrorEvents,"%g\n",CurrentSNR);
			}
		}

		/*---- Generating user bits for the current sector in NRZ format ----*/
//		printf("Running Temp code : Avoiding more than 3 transitions \n");
		for(i=0;i<UserSectorLength;i++)
		{
			UserBits[i] = (UniformRandomNumberGeneratorUserBits(&idum1)>0.5)?1:0;
			/* Temp Code : Avoid more than 3 transitions */
//			if((UserBits[i] - UserBits[i-1] != 0)&&(UserBits[i-1] - UserBits[i-2] != 0)&&(UserBits[i-2] - UserBits[i-3] != 0)&&(UserBits[i-3] - UserBits[i-4] != 0))
//				UserBits[i] = UserBits[i-1];
		}

/*		printf("Running Temp code : Converting tribits to single transitions \n");
		for(i=0;i<UserSectorLength;i++)
		{
			fpDebugFile = fopen(DebugFile,"a");
			fprintf(fpDebugFile, " %g",UserBits[i]);
			fclose(fpDebugFile,"\n");
		}
*/
		/* TEST CODE */
/*		printf("\nWARNING: main() --> Running Temporary Code. DELETE when no longer required \n");
		if(NumSectors >= 0)
		{
			for(i=0;i<SectorLength/2;i++)
				UserBits[i] = 0;
			for(i=SectorLength/2;i<SectorLength;i++)
				UserBits[i] = 1;
			UserBits[SectorLength/2 + 1] = 1;
		}
*/

		/*---- User bits are encoded by the encoder ----*/
		if(!UNCODED)
		{
			UserBits[0] = 0;
			//Enc_6By7MTRCode(UserBits,EncodedBits,&UserSectorLength);
			//Enc_8By9TMTRCode(UserBits,EncodedBits,&UserSectorLength);
			Enc_4By5RLLCode(UserBits,EncodedBits,&UserSectorLength);
		}
		if(DEBUG)
		{
			fpDebugFile = fopen(DebugFile,"a");
			for(i=1;i<UserSectorLength;i++)
			{
				fprintf(fpDebugFile, "%d",abs(UserBits[i]-UserBits[i-1]));
				if(i%4 == 0)
					fprintf(fpDebugFile, "  ");
			}
			fprintf(fpDebugFile,"\n");
			for(i=1;i<SectorLength;i++)
			{
				fprintf(fpDebugFile, "%d",abs(EncodedBits[i]-EncodedBits[i-1]));
				if(i%5 == 0)
					fprintf(fpDebugFile, " ");
			}
			fprintf(fpDebugFile,"\n");
			fclose(fpDebugFile);
		}			


		/* Hard coding for uncoded system */
		if(UNCODED)
		{
			for(i=0;i<SectorLength;i++)
			{
				EncodedBits[i] = UserBits[i];
			}
		}

		/*---- EncodedBits are pre-padded and post-padded with zeros ----*/
		/* 0 -> -1 and 1 -> +1	*/
		for(i=0;i<PaddedSectorLength;i++)
		{
			if(i < PrePaddingLength || i > (PrePaddingLength+SectorLength-1))
			{
				PaddedBits[i] = 0;
				PaddedBipolarBits[i] = 2*PaddedBits[i] - 1;		/* Should be 0 or +1/-1 ? */
			}else
			{
				PaddedBits[i] = EncodedBits[i-PrePaddingLength];
				PaddedBipolarBits[i] = 2*PaddedBits[i] - 1; 
			}
		}

		/*---- PaddedBits are oversampled and converted to Bipolar format ----*/
		/* 	0 -> -1 and 1 -> +1	*/
		/* This assignment will determine whether a given transition results in a +/- readback */
		for(i=0;i<OSSectorLength;i++)
		{
			OSBits[i] = PaddedBits[i/OSR];
			OSBipolarBits[i] = PaddedBipolarBits[i/OSR];
		}

		/*---- Encoded bits are transmitted through the channel ----*/
		/* Longitudinal or Perpendicular */
		/* Channel output is of length OSSectorLength + NumChannelTaps/2 */
		if(!strcmp(ChannelType,"Perpendicular") || !strcmp(ChannelType,"perpendicular") ||!strcmp(ChannelType,"Longitudinal") || !strcmp(ChannelType,"longitudinal"))
		{
			LongPerp(OSBipolarBits,OSSectorLength, ChannelOutput, ReadChannelCoeff,NumChannelTaps, TimeIndex, 
					ND, SigmaJitter, SigmaPulseBroad, OSR, ChannelType, j);

/*				if(DEBUG)
				{
					fpDebugFile = fopen(DebugFile,"a");
					for(i=0;i<OSSectorLength;i++)
					{
						//fprintf(fpDebugFile, " %g",OSBipolarBits[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<OSSectorLength;i++)
					{
						//fprintf(fpDebugFile, " %g",ChannelOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<201;i++)
					{
						fprintf(fpDebugFile, " %g",ReadChannelCoeff[i]);
					}
					fprintf(fpDebugFile,"\n");

					fclose(fpDebugFile);
				}
*/

		}else 
		/* HAMR */
		{
			/* Channel output is of length OSSectorLength only */
			/* Re-Initializing ChannelOutput */
			for(i=0;i<(OSSectorLength+ floor(NumChannelTaps/2)-1);i++)
			{
				ChannelOutput[i] = 0.0;
			}
			/* Non adaptive mode */
			if(j>=0)
			{
				Hamr(mp, pp, rp, N, OSBipolarBits, OSSectorLength, ChannelOutput, 
				OverSampledBitLength, OSR, NumSectors, K, Rho, &PW, X, XLength, PeakAmpPositionIndex, NLTSCompensation, 
				NumNLTSInfluencingTrans,TemperatureVariation, SigmaTempVariation, PeakTempTruncValue,
				WriteHdCrossTrackMov, MaxWriteHdCrTrMov, MeanWriteHdCrTrMov, SigmaJitter, HMD, SigmaHMDVariation, 
				NumARModelCoeff,ARModelCoeff, TemperatureModulation, ModulatedPeakTemp, &NormalizationFactor,0);
/*				if(DEBUG)
				{
					fpDebugFile = fopen(DebugFile,"a");
					for(i=0;i<OSSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",OSBipolarBits[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<OSSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",ChannelOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					fclose(fpDebugFile);
				}			
*/  
			}else
				/* Adaptive mode. So, avoiding media noise effects by making K,Rho and NumNLTSInfluencingTrans = 0 */
			{
				/* Channel output is of length OSSectorLength only */
				/* Re-Initializing ChannelOutput */
				for(i=0;i<(OSSectorLength+ floor(NumChannelTaps/2)-1);i++)
				{
					ChannelOutput[i] = 0.0;
				}

				Hamr(mp, pp, rp, N, OSBipolarBits, OSSectorLength, ChannelOutput, 
				OverSampledBitLength, OSR, NumSectors, KDummy, RhoDummy, &PW, X, XLength, PeakAmpPositionIndex, NLTSCompensation, 
				0,TemperatureVariation, SigmaTempVariation, PeakTempTruncValue,
				WriteHdCrossTrackMov, MaxWriteHdCrTrMov, MeanWriteHdCrTrMov, SigmaJitter, HMD, SigmaHMDVariation, 
				NumARModelCoeff,ARModelCoeff, TemperatureModulation, ModulatedPeakTemp, &NormalizationFactor,0);
			}
		}

		/*----- Adding electronic (AWGN) noise ----*/
		/* 
		SNR def: as given in page 6.7 of 'Coding and signal processing for magnetic rec sys'
		by Vasic and Kurtas. Since, we use bipolar inputs, Ei = 4(see pg 6.7).
		Therefore, in the calculation of noisesigma, we have an additional factor of 2.
		*/
		//printf("WARNING: main()-->Hard coding noise sigma calculation\n");
		//NoiseSigma = pow(OSR,0.5) * pow(10,-CurrentSNR/20) * 2 * pow(3/2.0,0.5);
 		NoiseSigma = pow(OSR,0.5) * pow(10,-CurrentSNR/20) * 2 * pow(ND/2.0,0.5);
		
		for (i=PrePaddingLength*OSR; i<OSSectorLength-PostPaddingLength*OSR; i++)
		{
			ChannelOutput[i]= ChannelOutput[i] + (NoiseSigma*GaussianRandomNumberGenerator(&idum2));
		}

		/*---- Channel output is Low-Pass Filtered ----*/
		/* LPF Output is of length [OSSectorLength + NumTaps - 1] */
		LPF(ChannelOutput, OSSectorLength, LPFOutput, (NumLPFTaps-1), (1/(double)OSR));
		//LPF(ChannelOutput, OSSectorLength, LPFOutput, (NumLPFTaps-1), (0.72/(double)OSR));

/*		if(DEBUG)
		{
			if(NumSectors >= -1)
			{
				fpDebugFile = fopen(DebugFile,"a");
				for(i=0;i<OSSectorLength;i++)
				{
					fprintf(fpDebugFile, " %g",OSBipolarBits[i]);
				}
				fprintf(fpDebugFile,"\n");
				for(i=0;i<OSSectorLength;i++)
				{
					//fprintf(fpDebugFile, " %g",ChannelOutput[i]);
				}
				fprintf(fpDebugFile,"\n");
				for(i=0;i<OSSectorLength;i++)
				{
					fprintf(fpDebugFile, " %g",LPFOutput[i]);
				}
				fprintf(fpDebugFile,"\n");
				fclose(fpDebugFile);
			}
		}		
*/	
		
		/*---- LPF output is downsampled ----*/
		for(i=0;i<PaddedSectorLength;i++)
		{
			DSOutput[i]=LPFOutput[i*OSR];
		}

		/*---- Downsampled output is Equalized to a fixed PR Target ----*/
		if(!strcmp(EqualizerType,"FixedPRTarget"))
		{/* PR */
			if(NumSectors < 0)
			{
				/* Adapting the equalizer filter coefficients to the PR target response */
				if(NumSectors == -NumEqSectors)
					StartFlag = 1;
				MSECheckFlag = AdaptEqualizer(PRImpRes, PRImpResLength, EqCoeff, NumEqTaps, PaddedBipolarBits, DSOutput, PaddedSectorLength, StartFlag);
				if(MSECheckFlag == 0) /* Implies that MSE is below threshold, so can stop adapting */
				{
					NumSectors = 0;	/* stop adapting */
				}else
				{
					NumSectors++;
				}
				continue;
			}else
			{
				/* Equalizing the current sector */
				NonCausalFIR(DSOutput, PaddedSectorLength, EqualizedOutput, EqCoeff, NumEqTaps);

				if(DEBUG)
				{
					fpDebugFile = fopen(DebugFile,"a");
					for(i=0;i<PaddedSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",PaddedBipolarBits[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<PaddedSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",DSOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<PaddedSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",EqualizedOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<NumEqTaps;i++)
					{
						fprintf(fpDebugFile, " %g",EqCoeff[i]);
					}
					fprintf(fpDebugFile,"\n");
					fclose(fpDebugFile);
				}

				/* Detecting input sequence for the current sector */
				if(!strcmp(DetectorType,"Viterbi"))
				{
					/*---- Equalized output is passed through a Viterbi Detector ----*/
					if(UNCODED)
					{
						//Viterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);					
						ClassicalViterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);					
						//ClassicalViterbiForTempMod(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength,
						//							States,ModulationValues,ValidIncomingTrans,&NumberOfStates);					
						//printf("Running test code: Viterbi_6By7MTRCode\n");
						//Viterbi_6By7MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);

					}else
					{
						//Viterbi_6By7MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);
						//Viterbi_8By9TMTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);
						ClassicalViterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, PRImpRes, PRImpResLength);					
					}
				}else if(!strcmp(DetectorType,"SOVA"))
				{
					if(UNCODED)
					{
						/*---- Equalized output is passed through a Soft-Output Viterbi (SOVA) Detector ----*/
						//SOVA(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, ViterbiSoftOutput, PRImpRes, PRImpResLength, NoiseSigma);
						//ClassicalSOVA(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, ViterbiSoftOutput, PRImpRes, PRImpResLength, NoiseSigma);
						SOVA_6By7MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, ViterbiSoftOutput, PRImpRes, PRImpResLength, NoiseSigma);
					}else
					{
						SOVA_8By9MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, ViterbiSoftOutput, PRImpRes, PRImpResLength, NoiseSigma);
					}
				}else
				{
					printf("\nERROR: Main -> Invalid Detector Type. Program terminated. \n");
					exit(-1);
				}
			}
			if(NumSectors == 0 && j==0)
			{
				fpSimulationSummary = fopen(SimulationSummary, "a");
				fprintf(fpSimulationSummary,"\n-------PR TARGET-------\n");
				fprintf(fpSimulationSummary,"Equalizer Type : %s\n","Fixed PR Target");
				fprintf(fpSimulationSummary,"PR Target :");
				for(i=0;i<PRImpResLength;i++)
					fprintf(fpSimulationSummary," %g",PRImpRes[i]);
				fprintf(fpSimulationSummary,"\n");
				fprintf(fpSimulationSummary,"Number of Equalizing Sectors = %d\n",NumEqSectors);
				fprintf(fpSimulationSummary,"Number of Equalizer Taps = %d\n",NumEqTaps);
				fclose(fpSimulationSummary);
			}
		}/* PR */

		/*---- Downsampled output is Equalized to a Generalized PR Target ----*/
		else if(!strcmp(EqualizerType,"GPRTarget"))
		{/* GPR */

/*			if(NumSectors<0)
			{
				printf("WARNING: Running temporary code\n");
				fpDebugFile = fopen("./GPREqualizer@ND3.375.dat","r");
				if(fpDebugFile==NULL)
					printf("ERROR: Cannot open file to read the equalizer target and coefficients\n");
				for(i=0;i<=GPRTargetLength-1;i++)
					fscanf(fpDebugFile,"%lg",&(GPRTarget[i]));
				for(i=0;i<=NumEqTaps-1;i++)
					fscanf(fpDebugFile,"%lg",&(EqCoeff[i]));
				fclose(fpDebugFile);
			}
*/
			/*---- LP Filtered and Downsampled output is Equalized to a Generalized PR Target ----*/
			if(NumSectors < 0)
			{
				/* Concatenate data (leaving out the pre and post paddiing) */
				for(i=0;i<SectorLength;i++)
				{
					GPRConcatChannelInput[(NumEqSectors+NumSectors)*SectorLength + i] = PaddedBipolarBits[i+PrePaddingLength];
					GPRConcatChannelOutput[(NumEqSectors+NumSectors)*SectorLength + i] = DSOutput[i+PrePaddingLength];
				}
				/* Find GPR target when all data has been concatenated */
				if(NumSectors == -1)
				{
					FindGPRTarget(GPRConcatChannelOutput, GPRConcatChannelInput, EqCoeff, GPRTarget, (NumEqSectors*SectorLength), GPRTargetLength, NumEqTaps);
				}
				NumSectors++;
				continue;
			}else
			{
				/* Equalizing the current sector */
				NonCausalFIR(DSOutput, PaddedSectorLength, EqualizedOutput, EqCoeff, NumEqTaps);

				if(DEBUG)
				{
					fpDebugFile = fopen(DebugFile,"a");
					for(i=0;i<PaddedSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",PaddedBipolarBits[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<PaddedSectorLength;i++)
					{
						//fprintf(fpDebugFile, " %g",DSOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<PaddedSectorLength;i++)
					{
						fprintf(fpDebugFile, " %g",EqualizedOutput[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<NumEqTaps;i++)
					{
						fprintf(fpDebugFile, " %g",EqCoeff[i]);
					}
					fprintf(fpDebugFile,"\n");
					for(i=0;i<GPRTargetLength;i++)
					{
						fprintf(fpDebugFile, " %g",GPRTarget[i]);
					}
					fprintf(fpDebugFile,"\n");
					fclose(fpDebugFile);
				}

				/* Detecting input sequence for the current sector */
				if(!strcmp(DetectorType,"Viterbi"))
				{
					/*---- Equalized output is passed through a Viterbi Detector ----*/
					if(UNCODED)
					{
						//Viterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
						ClassicalViterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
						//printf("Running test code: Viterbi_6By7MTRCode\n");
						//Viterbi_6By7MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
					}else
					{
						// Replace the last two function arguments according to the target type used
						//Viterbi_8By9TMTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
						//Viterbi_6By7MTRCode(ViterbiDelay, EqualizedOutput, PrePaddingLength, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
						ClassicalViterbi(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, GPRTarget, GPRTargetLength);
					}
				}else if(!strcmp(DetectorType,"SOVA"))
				{
					/*---- Equalized output is passed through a Soft-Output Viterbi (SOVA) Detector ----*/
					SOVA(ViterbiDelay, EqualizedOutput, PaddedSectorLength, ViterbiHardOutput, ViterbiSoftOutput, GPRTarget, GPRTargetLength, NoiseSigma);

				}else
				{
					printf("\nERROR: Main -> Invalid Detector Type. Program terminated. \n");
					exit(-1);
				}
			}
			if(NumSectors == 0 && j==0)
			{
				fpSimulationSummary = fopen(SimulationSummary, "a");
				fprintf(fpSimulationSummary,"\n-------PR TARGET-------\n");
				fprintf(fpSimulationSummary,"Equalizer Type : %s\n","GPR Target");
				fprintf(fpSimulationSummary,"GPR Target :");
				for(i=0;i<GPRTargetLength;i++)
					fprintf(fpSimulationSummary," %g",GPRTarget[i]);
				fprintf(fpSimulationSummary,"\n");
				fprintf(fpSimulationSummary,"Number of Equalizing Sectors = %d\n",NumEqSectors);
				fprintf(fpSimulationSummary,"Number of Equalizer Taps = %d\n",NumEqTaps);
				fclose(fpSimulationSummary);
			}
		}/* GPR */
		else
		{
			printf("\nERROR: Main -> Invalid Equalizer Type. Program terminated \n");
			exit(-1);
		}
		NumSectors++;

		/* TEST CODE : If temperature modulation is used, then detect isolated
		 * transitions with different PW50 (those that were initially tribits
		 */
		if(TemperatureModulation==1 && 0)
		{
			/* Hard coding for pre-determined values */
			PW1 = 69;
			PW2 = 94;
			Vp1 = 2*sqrt(2/PI);	/* Peak amplitude when PW50 = PW1 */
			Vp2 = 0.7322*Vp1;	/* Peak amplitude when PW50 = PW2 */
			NSymbols = 2;		/* Number of symbols to be considered for calc the MSE. Should be EVEN */
			NNeighbors = 20;	/* Number of bits around the isolated transition used to calculate the 
								   values at the 'NSymbols' positions around the isolated transition 
							       Shoule be EVEN */
			T = OverSampledBitLength*OSR;	/* Bit period */
			
			if(NumSectors == 1)
			{
				SampleValues = (double *)malloc(NSymbols * sizeof(double));
				ModulatedTribitIndex = (int *)malloc(PaddedSectorLength*sizeof(int));
			}
			for(i=0;i<=PaddedSectorLength-1;i++)
				ModulatedTribitIndex[i]=0;
			for(i=(NNeighbors/2);i<=PaddedSectorLength-(NNeighbors/2);i++)	/* TEMPORARY SOLN: Range of for loop needs to be changed */
			{
				/* Check for isolated transitions in the detected output */
				if((ViterbiHardOutput[i]-ViterbiHardOutput[i-1] != 0) && 
				  (ViterbiHardOutput[i-1]-ViterbiHardOutput[i-2] == 0) &&
				  (ViterbiHardOutput[i+1]-ViterbiHardOutput[i] == 0))
				{
					/* For a temperature modulated isolated transition, there should be no transition
					within 2T to the left of the transition
					*/
					if(ViterbiHardOutput[i-2]-ViterbiHardOutput[i-3] == 0)
					{
						/* From the down-sampled channel output, compare and see if this transition
						is of a different PW50. Consider the transition along with its 2 adjacent positions.
						*/
						/* Positions of the 'NNeighbors' bits: 
						 * -(NNeighbors/2)T	-(NNeighbors/2 - 1)T....-T 0 T 2T.....(NNeighbors/2)T
						 *
						 * Positions of the 'NSymbols' bits:
						 * -(NSymbols/2)T	-(NSymbols/2 - 1)T....-T 0 T 2T.....(NSymbols/2)T
						 */
						EnergyWithPW1 = 0;
						EnergyWithPW2 = 0;
						for(m=-(NSymbols/2);m<=(NSymbols/2);m++)
						{
							SampleValues[m+(NSymbols/2)]=0;
							for(n=-(NNeighbors/2);n<=(NNeighbors/2);n++)
							{
								if(n!=0)	/* Do not include the contribution of the isolated bit */
								{
									SampleValues[m+(NSymbols/2)] += (ViterbiHardOutput[i+n]-ViterbiHardOutput[i+n-1])*(Vp1/(1+pow(2*(m-n)*T/PW1,2)));
								}
							}
						}
						/* Contribution of the isolated transition */
						for(m=-(NSymbols/2);m<=(NSymbols/2);m++)
						{
							/* If the isolated transition has PW50 = PW1 */
							EnergyWithPW1 += pow(DSOutput[i+m] - (SampleValues[m+(NSymbols/2)] + (ViterbiHardOutput[i]-ViterbiHardOutput[i-1])*(Vp1/(1+pow(2*m*T/PW1,2)))) , 2);
							/* If the isolated transition has PW50 = PW2 */
							EnergyWithPW2 += pow(DSOutput[i+m] - (SampleValues[m+(NSymbols/2)] + (ViterbiHardOutput[i]-ViterbiHardOutput[i-1])*(Vp2/(1+pow(2*m*T/PW2,2)))) , 2);
						}
						if(EnergyWithPW2 < EnergyWithPW1) /* If true, then the actual PW50 = PW2 */
						{
							/* This isolated bit was originally a tribit */
							ModulatedTribitIndex[i] = 1;
						}
					}
				}
			}
			for(i=0;i<=PaddedSectorLength-1;i++)
			{
				if(ModulatedTribitIndex[i]==1)
				{
					ViterbiHardOutput[i] = abs(ViterbiHardOutput[i] - 1);	/* Invert bit */
					ViterbiHardOutput[i-1] = abs(ViterbiHardOutput[i-1] - 1);
				}
			}
		}
		if(NumSectors%10 == 0)
			printf("NumSectors = %d\n",NumSectors);

		/* Decoding */
		if(!UNCODED)
		{
			InvalidCW = Dec_4By5RLLCode(ViterbiHardOutput, DecodedOutput, PrePaddingLength, SectorLength);
			//Dec_6By7MTRCode(ViterbiHardOutput, DecodedOutput, PrePaddingLength, SectorLength);
			if(NumSectors%5 == 0)
				printf("NumSectors = %d\n",NumSectors);

			if(DEBUG && InvalidCW)
			{
				fpDebugFile = fopen(DebugFile,"a");
				for(i=1;i<UserSectorLength;i++)
				{
					fprintf(fpDebugFile, "%d",abs(UserBits[i]-UserBits[i-1]));
					if(i%4 == 0)
						fprintf(fpDebugFile, "  ");
				}
				fprintf(fpDebugFile,"\n");

				for(i=1;i<SectorLength;i++)
				{
					fprintf(fpDebugFile, "%d",abs(EncodedBits[i]-EncodedBits[i-1]));
					if(i%5 == 0)
						fprintf(fpDebugFile, " ");
				}
				fprintf(fpDebugFile,"\n");
				
				for(i=PrePaddingLength+1;i<SectorLength+PrePaddingLength;i++)
				{
					fprintf(fpDebugFile, "%d",abs(ViterbiHardOutput[i]-ViterbiHardOutput[i-1]));
					if((i-PrePaddingLength)%5 == 0)
						fprintf(fpDebugFile, " ");
				}
				fprintf(fpDebugFile,"\n");

				for(i=1;i<UserSectorLength;i++)
				{
					fprintf(fpDebugFile, "%d",abs(DecodedOutput[i]-DecodedOutput[i-1]));
					if(i%4 == 0)
						fprintf(fpDebugFile, "  ");
				}
				fprintf(fpDebugFile,"\n\n");
				fclose(fpDebugFile);
			}
		}

		/* Check for NRZI bit errors at the output of decoder */
		if(!UNCODED)
		{
			CountNRZIDecErr = 0;
			for(i=1;i<UserSectorLength;i++)
			{
				if(abs(DecodedOutput[i]-DecodedOutput[i-1]) != abs(UserBits[i]-UserBits[i-1]))
				{
					CountNRZIDecErr++;
				}
			}
			NumNRZIDecErr += CountNRZIDecErr;

			CountNRZDecErr = 0;
			for(i=0;i<UserSectorLength;i++)
			{
				if(DecodedOutput[i] != UserBits[i])
				{
					CountNRZDecErr++;
				}
			}
			NumNRZDecErr += CountNRZDecErr;
		}
		
		/* Detect NRZI bit errors at the output of Viterbi */
		CountNRZIVitErr = 0;
		for(i=PrePaddingLength+1;i<SectorLength+PrePaddingLength;i++)
		{
			if(abs(ViterbiHardOutput[i]-ViterbiHardOutput[i-1]) != abs(PaddedBits[i]-PaddedBits[i-1]))
			{
				CountNRZIVitErr++;
			}
		}
		NumNRZIVitErr += CountNRZIVitErr;
		
		/* Detect NRZ bit errors at the output of Viterbi */
		CountNRZVitErr = 0;
		for(i=PrePaddingLength;i<SectorLength+PrePaddingLength;i++)
		{
			if(ViterbiHardOutput[i] != PaddedBits[i])
			{
				CountNRZVitErr++;
			}
		}
		NumNRZVitErr += CountNRZVitErr;

		/* Assuming that an NRZI error would also lead to atleast one NRZ error and vice-versa */
		if(CountNRZIDecErr > 0)
			NumDecSecErr++;
		if(CountNRZIVitErr > 0)
			NumVitSecErr++;


		if(ERROR_EVENTS)
		{
//			if(NumSectors==1)
//				printf("\nWARNING: Running test code\n");
			if(CountNRZVitErr > 0)
			{
				/* Avoiding a space in the beginning of the file, since it creates some problems in matlab */
				fprintf(fpErrorEvents,"%+d",ViterbiHardOutput[PrePaddingLength] - PaddedBits[PrePaddingLength]);
				for(i=PrePaddingLength+1;i<SectorLength+PrePaddingLength;i++)
				{
					fprintf(fpErrorEvents," %+d",ViterbiHardOutput[i] - PaddedBits[i]);
				}
				fprintf(fpErrorEvents,"\n");
				fprintf(fpErrorEvents,"%+d",PaddedBits[PrePaddingLength]);
				for(i=PrePaddingLength+1;i<SectorLength+PrePaddingLength;i++)
				{
					fprintf(fpErrorEvents," %+d",PaddedBits[i]);
				}
				fprintf(fpErrorEvents,"\n");
				fprintf(fpErrorEvents,"%+d",ViterbiHardOutput[PrePaddingLength]);
				for(i=PrePaddingLength+1;i<SectorLength+PrePaddingLength;i++)
				{
					fprintf(fpErrorEvents," %+d",ViterbiHardOutput[i]);
				}
				fprintf(fpErrorEvents,"\n\n");

				/*for(i=PrePaddingLength;i<SectorLength+PrePaddingLength;i++)
				{
					fprintf(fpErrorEvents, "%g\n",ViterbiSoftOutput[i]);
				}
				fprintf(fpErrorEvents,"\n");
				*/
			}
		}
	}/* sector count loop */


if(j==0)
{
	fpSimulationSummary = fopen(SimulationSummary, "a");
	fprintf(fpSimulationSummary,"\n-------SIMULATION RESULTS-------\n");
	if(UNCODED)
	{
		fprintf(fpSimulationSummary, "<CurrentSNR>  <BER(NRZ/DEC)>  <BER(NRZI/DEC)>   <SER>   <Number Of Sectors>   <NRZ Bit Errors>  <NRZI Bit Errors>   <Sector Errors>\n");
	}else
	{
		fprintf(fpSimulationSummary, "<CurrentSNR>  <BER(NRZ/DEC)>  <BER(NRZI/DEC)>   <BER(NRZ/VIT)>   <BER(NRZI/VIT)>   <Dec SER>   <Vit SER>   <Number Of Sectors>   <NRZ Dec Bit Errors>  <NRZI Dec Bit Errors>   <NRZ Vit Bit Errors>   <NRZI Vit Bit Errors>   <Dec Sector Errors>  <Vit Sector Errors>\n");
	}
	fclose(fpSimulationSummary);
}

/* Calculate BER and SER */
fpSimulationSummary = fopen(SimulationSummary, "a");
//ber[j] = (double)NumBitErr/(double)(NumSectors*(UserSectorLength-1)); /* For NRZI BER in MTR codes */
//ber[j] = (double)NumBitErr/(double)(NumSectors*SectorLength);
ser[j] = (double)NumVitSecErr/(double)NumSectors;
//printf("CurrentSNR = %-1.5e BER = %-1.5e SER = %-1.5e NumSec = %7d NumBitErr = %7d NumSecErr = %7d\n",CurrentSNR,ber[j],ser[j],NumSectors,NumBitErr,NumSecErr);
//fprintf(fpSimulationSummary, "%-1.5e %-1.5e %-1.5e %7d %7d %7d\n",CurrentSNR,ber[j],ser[j],NumSectors,NumBitErr,NumSecErr);
//fprintf(fpSimulationSummary, "%-1.5e %-1.5e %-1.5e %7d %7d %7d %-1.5e\n",CurrentSNR,ber[j],ser[j],NumSectors,NumBitErr,NumSecErr,(double)NumViterbiBitErr/(double)(NumSectors*(SectorLength-1)));
if(UNCODED)
{
	printf("CurrentSNR=%-1.5e BER(NRZ/VIT)=%-1.5e BER(NRZI/VIT)=%-1.5e SER=%-1.5e NumSec=%7d NumNRZBitErr=%7d NumNRZIBitErr=%7d NumSecErr=%7d\n\n",CurrentSNR, (double)NumNRZVitErr/(double)(NumSectors*SectorLength),
		   (double)NumNRZIVitErr/(double)(NumSectors*(SectorLength-1)),ser[j],NumSectors,NumNRZVitErr,NumNRZIVitErr,NumVitSecErr);
	fprintf(fpSimulationSummary,"%-1.5e %-1.5e %-1.5e %-1.5e %7d %7d %7d %7d\n",CurrentSNR, (double)NumNRZVitErr/(double)(NumSectors*SectorLength),
		   (double)NumNRZIVitErr/(double)(NumSectors*(SectorLength-1)),ser[j],NumSectors,NumNRZVitErr,NumNRZIVitErr,NumVitSecErr);
}else
{
	printf("CurrentSNR=%-1.5e BER(NRZ/DEC)=%-1.5e BER(NRZI/DEC)=%-1.5e BER(NRZ/VIT)=%-1.5e BER(NRZI/VIT)=%-1.5e SER(Dec)=%-1.5e SER(Vit)=%-1.5e NumSec=%7d NumNRZDecErr = %7d NumNRZIDecErr = %7d NumNRZVitErr = %7d NumNRZIVitErr = %7d NumDecSecErr=%7d NumVitSecErr=%7d\n\n",
		CurrentSNR, (double)NumNRZDecErr/(double)(NumSectors*UserSectorLength),(double)NumNRZIDecErr/(double)(NumSectors*(UserSectorLength-1)),(double)NumNRZVitErr/(double)(NumSectors*SectorLength),
		(double)NumNRZIVitErr/(double)(NumSectors*(SectorLength-1)), (double)NumDecSecErr/(double)NumSectors, ser[j], NumSectors,NumNRZDecErr,NumNRZIDecErr, NumNRZVitErr, NumNRZIVitErr, NumDecSecErr, NumVitSecErr);
	fprintf(fpSimulationSummary,"%-1.5e %-1.5e %-1.5e %-1.5e %-1.5e %-1.5e %-1.5e %7d %7d %7d %7d %7d %7d %7d\n",CurrentSNR, 
		(double)NumNRZDecErr/(double)(NumSectors*UserSectorLength),(double)NumNRZIDecErr/(double)(NumSectors*(UserSectorLength-1)),(double)NumNRZVitErr/(double)(NumSectors*SectorLength),
		(double)NumNRZIVitErr/(double)(NumSectors*(SectorLength-1)), (double)NumDecSecErr/(double)NumSectors, ser[j], NumSectors,NumNRZDecErr,NumNRZIDecErr, NumNRZVitErr, NumNRZIVitErr, NumDecSecErr,NumVitSecErr);
}
fclose(fpSimulationSummary);

}/* Loops over all SNR values */

fpSimulationSummary = fopen(SimulationSummary, "a");
fprintf(fpSimulationSummary,"\nTime elapsed: %g m\n", (((double)clock() - start) / CLOCKS_PER_SEC)/60.0);
fclose(fpSimulationSummary);
if(ERROR_EVENTS)
	fclose(fpErrorEvents);

/* Freeing Memory */
free(mp);free(pp);free(rp);
free(K);free(Rho);
free(UserBits);
free(PaddedBits);
free(PaddedBipolarBits);
free(OSBits);
free(OSBipolarBits);
free(ChannelOutput);
free(LPFOutput);
free(DSOutput);
free(EqualizedOutput);
free(ViterbiHardOutput);
free(ViterbiSoftOutput);
free(DecodedOutput);
free(ReadChannelCoeff);
free(TimeIndex);
free(EqCoeff);
free(GPRTarget);
free(GPRConcatChannelInput);
free(GPRConcatChannelOutput);
free(ber);
free(ser);
if(!strcmp(ChannelType,"Hamr") || !strcmp(ChannelType,"hamr") || !strcmp(ChannelType,"HAMR"))
{
	free(PeakAmpPositionIndex);
	free(X);
	free(KDummy);
	free(RhoDummy);
}
//printf("WARNING: main()--> Running temporary code\n");
free(States);
free(ModulationValues);
for(i=0;i<1024;i++)
	free(ValidIncomingTrans[i]);

}

